#include "SD.h"

#define STACKSIZE 512
#define MINMEM (15*1024)
#define ALLTABLES (512+512+512+256)

ULONG	CompMask, /* What to signal to Copier when the compression/decompression is finished */
		CompCoMask;	/* What to signal back to the copy task */

char	FirstRound;	/* This is used only by comp.a */
short	CurrentBlock;	/* Idem */
short CurBuf = -1;	/* The buffer we are filling while compressing; -1 means no one available */

unsigned long DMACheck[8];	/* Checksums for the DMA buffer */

static unsigned char *RAMBuffer[80];	/* pointers to the 80 buffer chunks */
static unsigned long Check[160];			/* Checksums of the internal RAM buffer */

static const	ULONG	SYNC = 0x4489;
static const	ULONG	SYNCSYNC = 0x44894489;
static const	ULONG	MFMZeroes = 0xAAAAAAAA;
static const	ULONG	MFMMask = 0x55555555;

static unsigned char BufferName[] = "SD_Buffer",
							TaskName[] = "SD_Pack",
							Name[31], 								/* The disk name */
							VDiskOpen,								/* Is a virtual disk open? */
							HDFile,									/* Do we have to delete the HD file? */
							Compressing, 							/* Are we compressing? */
							TaskAlive, 								/* Is the compression task alive? */
							NotCompressed[10],					/* Which tracks haven't been compressed? */
							*CurPos, 								/* Current position while compressing */
							*Stack,									/* Stack of the compression task */
							*PackingTable,
							*UnpackingTable,						/* Tables for compression */
							*BlockPointers[2][SPT*2];			/* Pointers to the disk block data for
																			this compression pass */
static unsigned char ReadID[] = "Read",
							VerifyID[] = "Verify",
							UnitTrackMsg[] = "%s error on unit %d, track %d: %s\n",
							NoFirstSector[] = "Can't find first sector after gap",
							WrongSecNumber[] = "Wrong sector number",
							WrongTrackNumber[] = "Wrong track number",
							WrongNextNumber[] = "Wrong number of sectors before gap",
							BadChecksum[] = "Bad checksum",
							VerifyError[] = "Failed verify",
							ErrorName[] = "<BAD NAME>";

static unsigned short	BufNumber,		/* The buffers we got */
								CompCyl;			/* The cylinder we are compressing */
static unsigned short *Count, *Count2;	/* Temporary tables */
static struct Task Task;					/* The compression task */
static struct IOStdReq *VIOStdReq;		/* The IOReq for the virtual disk */
static struct MsgPort *VPort;
static ULONG DosMark;						/* The DOS mark of the current disk */
static long h;									/* The handle for the HD file buffer */

static unsigned char ConsoleName[] = "CON:0/130/600/70/SuperDuper Error Report";
static long ErrorConsole;

static void PrintError(unsigned char *Operation, unsigned int Unit, unsigned int Track, unsigned char *ErrorMsg) {

	if (!(PrintErrors && (stdout || ErrorConsole))) return;
	ObtainSemaphore(&Semaphore);
	sprintf(s, UnitTrackMsg, Operation, Unit, Track, ErrorMsg);
	Write(stdout ? stdout : ErrorConsole, s, strlen(s));
	ReleaseSemaphore(&Semaphore);
}

void OpenErrorConsole(void) {
	if (!ErrorConsole && !stdout) ErrorConsole = Open(ConsoleName, MODE_NEWFILE);
}

void CloseErrorConsole(void) {
	if (ErrorConsole) Close(ErrorConsole);
	ErrorConsole = NULL;
}

/* This is the task running while we compress */

static void __saveds Compress(void) {

	short i;
	unsigned long len, freespace;
	signed char CompCoSignal;

	CompCoMask = MASK(CompCoSignal = AllocSignal(-1));

	do {
		if (CurBuf == BufNumber) {
			CurBuf = -1;
			goto SignalCopier;
		}
		SetTable(BlockPointers[CompCyl%2], Count);
		CopyMemQuick(Count, Count2, 512);
		MakeIndexTable(Count2, UnpackingTable, (unsigned short *)PackingTable);
		BuildTable(UnpackingTable, PackingTable);
		len = PreviewLength(Count, UnpackingTable);
		freespace = RAMBUFFERBLOCK-(CurPos-RAMBuffer[CurBuf]);
		if (len+128>=TRACKSIZE*2) {
			if (++CurBuf == BufNumber && freespace != RAMBUFFERBLOCK) {
				CurBuf = -1;
				goto SignalCopier;
			}
			for(i=0; i<freespace/TD_SECTOR; i++) {
				if ((unsigned long)(BlockPointers[CompCyl%2][i])%4)
					CopyMem(BlockPointers[CompCyl%2][i], CurPos, TD_SECTOR);
				else CopyMemQuick(BlockPointers[CompCyl%2][i], CurPos, TD_SECTOR);
				CurPos += TD_SECTOR;
			}
			if (freespace == RAMBUFFERBLOCK) CurPos = RAMBuffer[CurBuf];
			else {
				if (freespace %= TD_SECTOR) CopyMem(BlockPointers[CompCyl%2][i], CurPos, freespace);
				CopyMem(BlockPointers[CompCyl%2][i]+freespace, RAMBuffer[CurBuf], TD_SECTOR-freespace);
				CurPos = RAMBuffer[CurBuf]+TD_SECTOR-freespace;
				for(i++; i<SPT*2; i++) {
					if ((unsigned long)(BlockPointers[CompCyl%2][i])%4)
						CopyMem(BlockPointers[CompCyl%2][i], CurPos, TD_SECTOR);
					else CopyMemQuick(BlockPointers[CompCyl%2][i], CurPos, TD_SECTOR);
					CurPos += TD_SECTOR;
				}
			}
			NotCompressed[CompCyl/8] |= (1<<(CompCyl%8));
		}
		else {
			if (freespace < len+128 && CurBuf == BufNumber-1) {
				CurBuf = -1;
				goto SignalCopier;
			}
			ConvertTable(UnpackingTable, PackingTable);
			if (128 >= freespace) {
				CopyMemQuick(UnpackingTable, CurPos, freespace);
				CurPos = RAMBuffer[++CurBuf]+128-freespace;
				if (freespace != 128) CopyMemQuick(UnpackingTable+freespace, RAMBuffer[CurBuf], 128-freespace);
				freespace = RAMBUFFERBLOCK-(128-freespace);
			}
			else {
				CopyMemQuick(UnpackingTable, CurPos, 128);
				CurPos += 128;
				freespace -= 128;
			}
			if (len>=freespace) {
				Pack(BlockPointers[CompCyl%2], CurPos, PackingTable, freespace, RAMBuffer[++CurBuf]);
				CurPos = RAMBuffer[CurBuf]+len-freespace;
			}
			else {
				Pack(BlockPointers[CompCyl%2], CurPos, PackingTable, len+4, RAMBuffer[CurBuf]);
				CurPos += len;
			}
		}

	SignalCopier:
		Signal(Copier, CompMask);
		Wait(CompCoMask);
	} while(CurBuf != -1);

	FreeSignal(CompCoSignal);
}

/* This is the task running while we decompress */

static void __saveds Decompress(void) {

	short i;
	unsigned long freespace;
	signed char CompCoSignal;

	CompCoMask = MASK(CompCoSignal = AllocSignal(-1));

	do {
		freespace = RAMBUFFERBLOCK-(CurPos-RAMBuffer[CurBuf]);
		if (NotCompressed[CompCyl/8] & (1<<(CompCyl%8))) {
			if (freespace == RAMBUFFERBLOCK) {
				for(i=0; i<SPT; i++) CopyMemQuick(CurPos+i*TD_SECTOR, BlockPointers[CompCyl%2][i], TD_SECTOR);
				Signal(Copier, CompMask);
				Wait(CompCoMask);
				if (CurBuf == -1) goto FreeAndExit;
				for(i=SPT; i<SPT*2; i++) CopyMemQuick(CurPos+i*TD_SECTOR, BlockPointers[CompCyl%2][i], TD_SECTOR);
				CurPos = RAMBuffer[++CurBuf];
			}
			else if (freespace >= SPT*TD_SECTOR) {
				for(i=0; i<SPT; i++) CopyMemQuick(CurPos+i*TD_SECTOR, BlockPointers[CompCyl%2][i], TD_SECTOR);
				Signal(Copier, CompMask);
				Wait(CompCoMask);
				if (CurBuf == -1) goto FreeAndExit;
				CurPos += SPT*TD_SECTOR;
				freespace -= SPT*TD_SECTOR;
				for(i=0; i<freespace/TD_SECTOR; i++) CopyMemQuick(CurPos+i*TD_SECTOR, BlockPointers[CompCyl%2][i+SPT], TD_SECTOR);
				CurPos += i*TD_SECTOR;
				if (freespace %= TD_SECTOR) CopyMemQuick(CurPos, BlockPointers[CompCyl%2][i+SPT], freespace);
				CopyMemQuick(RAMBuffer[++CurBuf], BlockPointers[CompCyl%2][i+SPT]+freespace, TD_SECTOR-freespace);
				CurPos = RAMBuffer[CurBuf]+TD_SECTOR-freespace;
				for(i++; i<SPT; i++) {
					CopyMemQuick(CurPos, BlockPointers[CompCyl%2][i+SPT], TD_SECTOR);
					CurPos += TD_SECTOR;
				}
			}
			else {
				for(i=0; i<freespace/TD_SECTOR; i++) CopyMemQuick(CurPos+i*TD_SECTOR, BlockPointers[CompCyl%2][i], TD_SECTOR);
				CurPos += i*TD_SECTOR;
				if (freespace %= TD_SECTOR) CopyMemQuick(CurPos, BlockPointers[CompCyl%2][i], freespace);
				CopyMemQuick(RAMBuffer[++CurBuf], BlockPointers[CompCyl%2][i]+freespace, TD_SECTOR-freespace);
				CurPos = RAMBuffer[CurBuf]+TD_SECTOR-freespace;
				for(i++; i<SPT; i++) {
					CopyMemQuick(CurPos, BlockPointers[CompCyl%2][i], TD_SECTOR);
					CurPos += TD_SECTOR;
				}
				Signal(Copier, CompMask);
				Wait(CompCoMask);
				if (CurBuf == -1) goto FreeAndExit;
				for(i=0; i<SPT; i++) {
					CopyMemQuick(CurPos, BlockPointers[CompCyl%2][SPT+i], TD_SECTOR);
					CurPos += TD_SECTOR;
				}
			}
		}
		else {
			if (freespace>128) {
				CopyMemQuick(CurPos, UnpackingTable, 128);
				BuildTable(UnpackingTable, PackingTable);
				CurPos = UnPack(CurPos+128, BlockPointers[CompCyl%2], UnpackingTable, freespace-128, RAMBuffer[CurBuf+1]);
				if (CurBuf == -1) goto FreeAndExit;
				if (CurPos-RAMBuffer[CurBuf+1]<RAMBUFFERBLOCK && CurPos-RAMBuffer[CurBuf+1]>=0) CurBuf++;
			}
			else {
				CopyMemQuick(CurPos, UnpackingTable, freespace);
				if (freespace != 128) CopyMemQuick(RAMBuffer[CurBuf+1], UnpackingTable+freespace, 128-freespace);
				BuildTable(UnpackingTable, PackingTable);
				CurBuf++;
				CurPos = UnPack(RAMBuffer[CurBuf]+128-freespace, BlockPointers[CompCyl%2], UnpackingTable, RAMBUFFERBLOCK-128+freespace, RAMBuffer[CurBuf]);
				if (CurBuf == -1) goto FreeAndExit;
			}
		}

		Signal(Copier, CompMask);
		Wait(CompCoMask);
	} while(CurBuf != -1);

FreeAndExit:
	FreeSignal(CompCoSignal);
}

/* Reset compression variables. Call it from anywhere. */

void SetUpCompression(unsigned char Reading) {
	CurBuf = 0;
	CurPos = RAMBuffer[0];
	if (Reading) setmem(NotCompressed, sizeof(NotCompressed), 0);
}


/* Here we start the comp/decomp task. If it's still alive, we simply
	signal it. All the following functions up to the Alloc/Free series
	can only be called from the copy task. */

static void StartTask(void Code()) {

	if (TaskAlive) {
		Signal(&Task, CompCoMask);
	}
	else {
		setmem(&Task, sizeof(Task), 0);
		Task.tc_Node.ln_Type = NT_TASK;
		Task.tc_Node.ln_Name = TaskName;
		Task.tc_SPLower = (void *)Stack;
		Task.tc_SPReg = Task.tc_SPUpper = (void *)(Stack+STACKSIZE);
		AddTask(&Task, (void *)Code, NULL);
		TaskAlive = 1;
	}
	Compressing = 1;
}


/* We start the comp/decomp */

void StartCompression(unsigned short Track) {
	CompCyl = Track/2;
	StartTask(Compress);
}


void StartDecompression(unsigned short Track, unsigned char *B1, unsigned char *B2) {
	short i;
	CompCyl = Track/2;
	for(i=0; i<SPT; i++) BlockPointers[CompCyl%2][i] = B1+SE_Data+TD_SECTOR+i*BLOCKSIZE;
	for(i=0; i<SPT; i++) BlockPointers[CompCyl%2][i+SPT] = B2+SE_Data+TD_SECTOR+i*BLOCKSIZE;
	StartTask(Decompress);
}

/* We wait for the end of compression. We get a non-zero result if compression
	is over for lack of space. */

unsigned char WaitCompression(unsigned char Rem) {

	if (Compressing) {
		Wait(CompMask);
		Compressing = 0;
		if (Rem || CurBuf == -1) {
			Rem = (CurBuf == -1);
			TaskAlive = 0;
			CurBuf = -1;
			Signal(&Task, CompCoMask);
			return(Rem);
		}
	}
	return(0);
}

/* This routine is used to sync with the two-pass decompression (one track
	at a time). */

void WaitInterDecompression(void) {

	if (Compressing) {
		Wait(CompMask);
		Signal(&Task, CompCoMask);
	}
}


/* All the following allocation/deallocation functions should be called from the main task. */


/*	This routine frees RAM buffers and compression tables. */

static void FreeBuffers(void) {

	register short i;

	for(i=0; i<80; i++)
		if (RAMBuffer[i]) {
			FreeMem(RAMBuffer[i], RAMBUFFERBLOCK);
			RAMBuffer[i] = NULL;
		}
	BufNumber = 0;
	Compression = 0;
	if (Count) {
		FreeMem(Count, ALLTABLES+STACKSIZE);
		Count = NULL;
	}
}

/* Here we allocate buffers and, if we don't get enough, we allocate compression stuff. */

unsigned char AllocBuffers(void) {

	register short i;
	register unsigned char *p;

	Compression = 0;
	FlushMem();

	for (i=0; i<80; i++)
		if (AvailMem(MEMF_PUBLIC)<MINMEM || (RAMBuffer[i]=AllocMem(RAMBUFFERBLOCK, MEMF_PUBLIC)) == NULL) {
			if (i<20) {
				FreeBuffers();
				return(0);
			}
			else break;
		}
		else BufNumber++;

	if (BufNumber == 80) return(1);

	if ((p = AllocMem(ALLTABLES+STACKSIZE, MEMF_PUBLIC)) == NULL) {
		FreeBuffers();
		return(0);
	}

	Count = (unsigned short *)p;
	Count2 = Count+256;
	PackingTable = (unsigned char *)(Count2+256);
	UnpackingTable = PackingTable+512;
	Stack = UnpackingTable+256;
	Compression = 1;
	ObtainSemaphore(&Semaphore);
	sprintf(s, "Got %d buffers. Compression is on.", BufNumber);
	Acknowledge(s);
	ReleaseSemaphore(&Semaphore);
	return(1);
}



unsigned char CopyTaskOpenHDBuffer(void) {
	return((unsigned char)((h = Open(BufferName, MODE_OLDFILE)) != NULL));
}

void CopyTaskCloseHDBuffer(void) {
	if (h) {
		Close(h);
		h = NULL;
	}
}

static void FreeHDBuffer(void) {
	if (HDFile) {
		DeleteFile(BufferName);
		HDFile = FALSE;
	}
}

/* Here we allocate an image file on the hard disk. We open the file in MODE_READWRITE,
	so interested people can copy it. */

unsigned char AllocHDBuffer(void) {

	int i;
	register struct InfoData *InfoData;
	long h = 0;
	BPTR l;

	if (l = Lock(BufferName, ACCESS_READ)) {
		UnLock(l);
		return(1);
	}

	if ((InfoData = AllocMem(sizeof(struct InfoData), MEMF_PUBLIC | MEMF_CLEAR)) &&
		Info(((struct Process *)Me)->pr_CurrentDir, InfoData) &&
		InfoData->id_NumBlocks-InfoData->id_NumBlocksUsed > 1786L &&
		(h = Open(BufferName, MODE_NEWFILE))) {

		for(i=0; i<80; i++)
			if (Write(h, VerifyBuffer[0], SPT*TD_SECTOR*2)<SPT*TD_SECTOR*2) {
				Close(h);
				h = 0;
				DeleteFile(BufferName);
				break;
			}
		if (h) Close(h);
	}

	if (InfoData) FreeMem(InfoData, sizeof(struct InfoData));
	return(HDFile = (unsigned char)(h != 0));
}

unsigned char CopyTaskOpenVDisk(void) {

	if ((VPort = CreatePort(NULL, 0)) &&
		(VIOStdReq = CreateStdIO(VPort)) &&
		(VDiskOpen = !OpenDevice(GetDeviceName(), GetVDNumber(), VIOStdReq, 0))) return(1);
	else {
		CopyTaskCloseVDisk();
		return(0);
	}
}

void CopyTaskCloseVDisk(void) {

	if (VDiskOpen) {
		CloseDevice(VIOStdReq);
		VDiskOpen = 0;
	}
	if	(VIOStdReq) {
		DeleteStdIO(VIOStdReq);
		VIOStdReq = NULL;
	}
	if	(VPort) {
		DeletePort(VPort);
		VPort = NULL;
	}
}

/* Here we open any exec device and check if it has enough blocks. Then we close. */

unsigned char AllocVDisk(void) {

	unsigned char result = 0;
	struct IOStdReq *IOStdReq;

	if (!GetDeviceName()) return(0);
	if (IOStdReq = CreateStdIO(TPort)) {
		if (!OpenDevice(GetDeviceName(), GetVDNumber(), IOStdReq, 0)) {
			IOStdReq->io_Command = CMD_READ;
			IOStdReq->io_Length = TD_SECTOR;
			IOStdReq->io_Data = VerifyBuffer[0]+128;
			IOStdReq->io_Offset = (SPT*TD_SECTOR)*160UL-TD_SECTOR;
			if (!DoIO(IOStdReq)) result = 1;
			CloseDevice(IOStdReq);
		}
		DeleteStdIO(IOStdReq);
	}
	return(result);
}

void FreeAllBuffers(void) {
	FreeBuffers();
	FreeHDBuffer();
}



/* The following functions are just computations, and can be called
	from everywhere. From StoreTrack afterwards, the blitter is used,
	so you have to call from the copy task. */


/*	This routine recalculates the checksum for a block, and updates it in
	word 5.  The sum of all the words in a block must be 0. */

static void ReCheck(long *w) {

	register short i;
	register long CheckSum;

	CheckSum = 0;
	for (i=0; i<128; i++)
		CheckSum += w[i];
	w[5] -= CheckSum;
}

/* We simply DateStamp the creation date, modification date, and
	rechecksum the block. A little random factor is added since we
	update many disks in a short period of time. We also increment the name of
	the disk if required. */

static void UpdateRootBlock(unsigned char *b, unsigned char Date, unsigned char IncName) {

	static unsigned char ShiftFactor = 1;

	if (Date){
		ShiftFactor = (ShiftFactor+2)%50;
		DateStamp((void *)(b + 420));
		DateStamp((void *)(b + 484));
		((struct DateStamp *)(b + 420))->ds_Tick = (((struct DateStamp *)(b + 420))->ds_Tick+ShiftFactor)%3000;
		((struct DateStamp *)(b + 484))->ds_Tick = (((struct DateStamp *)(b + 484))->ds_Tick+ShiftFactor)%3000;
	}
	if (IncName) {
		IncrementDiskName();
		strcpy((char *)(b+433), GetDiskName());
		*(b+432) = strlen(b+433);
	}
	ReCheck((void *)b);
}

/*
 *	  This routine creates data for the format option.	 Note the clever
 *	  way the data is built up; this routine should build a disk exactly
 *	  the same way the standard AmigaDOS format does.	If we are on track
 *	  40, some additional work must be done to create a root block and
 *	  bitmap, but it's not too bad. I don't know if this is the right
 *   way to do under FFS, but since it works . . .
 */
static void MakeFormatData(unsigned short Sector, unsigned char *b, unsigned char UseFFS) {

	register unsigned long *p, cs, i;
	register unsigned char *q;

	p = (unsigned long *)b;
	if (Sector != 880 && Sector != 881) {
		cs = (((ULONG)('D') << 24) | ((ULONG)('O') << 16) | ((ULONG)('S') << 8) | (UseFFS ? '\1' : '\0')) + (((long)(Sector/(SPT*2)) & 48) << 16);
		for (i=0; i<TD_SECTOR/4; i++)
			*p++ = cs + (i & 3327);
		return;
	}
	setmem(b, TD_SECTOR, 0);
	if (Sector == 880) {
		p[0] = 2;
		p[3] = 0x48;
		p[78] = 0xffffffff;
		p[79] = 0x371;
		q = (unsigned char *)(p + 108);
		strcpy((char *)(q+1), GetDiskName());
		*q = strlen(q+1);
		p[127] = 1;
		UpdateRootBlock(b, TRUE, FALSE);
	}
	else {
		for (i=1; i<55; i++)
			p[i] = 0xffffffffUL;
		p[0] = 0xc000c037UL;
		p[28] = 0xffff3fffUL;
		p[55] = 0x3fffffffUL;
	}
}


/* Various obvious service functions */

static void SetBufferStart(unsigned char *Buffer) {
	*((ULONG *)(Buffer)) = MFMZeroes;
	*((ULONG *)(Buffer+SE_Sync)) = SYNCSYNC;
}


static void SetBufferBorders(unsigned char *Buffer) {
	*((ULONG *)(Buffer-4)) = *((ULONG *)(Buffer-8)) = *((ULONG *)(Buffer+SPT*BLOCKSIZE)) = MFMZeroes;
}

signed char FindOffset(unsigned char *Buffer) {
	if (*((ULONG *)(Buffer)) == SYNCSYNC) return(-4);
	else if (*((USHORT *)(Buffer)) == SYNC) return(-6);
	else return(-8);
}

/* Calculates the gap length of the given Buffer (not offset). If anything
	is wrong, GAPSIZE is returned. */

unsigned short GapLength(unsigned char *Buffer) {

	unsigned char NumSec;
	unsigned char *p;

	NumSec = (unsigned char)(DecodeLong((Buffer+=FindOffset(Buffer))+SE_Header));
	if (NumSec<1 || NumSec>SPT) {
		return(GAPSIZE);
	}
	Buffer += NumSec*BLOCKSIZE;
/*	p = FindSync(Buffer, GAPSIZE);
	if (p) printf("Gap: %d\n",(unsigned short)(max(p-Buffer-4,MINGAP)+GAPADD));
	else printf("No gap!\n");*/
	if (p = FindSync(Buffer, GAPSIZE)) return((unsigned short)(max(p-Buffer-4,MINGAP)+GAPADD));
	else return(GAPSIZE);
}


/* We decode a track and we store it. If ReallyStore is zero we simply check it. From
	here, we're using the blitter, so you have to call from the copy task.
	We return 0 if everything is OK, 1 on error. */

signed char StoreTrack(unsigned char *p, unsigned char Unit, unsigned short Track, unsigned char ReallyStore, unsigned char Date) {

	register unsigned char i;
	unsigned char Sec, NumSec;
	signed char Result = 0;

	if (Track>159) return(1);
	if (Track == 0) DosMark = 0;

	p += FindOffset(p);
	NumSec = (unsigned char)(DecodeLong(p+SE_Header));
	if ((unsigned char)(DecodeLong(p+SE_Header)>>16) != Track) {
		PrintError(ReadID, Unit, Track, WrongTrackNumber);
		Result = 1;
	}

	if (NumSec>0 && NumSec<=SPT)
		for(i=0; i<SPT; i++) {
			if (i == NumSec) {
				if ((p = FindSync(p, GAPSIZE)) == NULL) {
					PrintError(ReadID, Unit, Track, NoFirstSector);
					Result = 1;
					break;
				}
				p += FindOffset(p);
			}
			Sec = (unsigned char)(DecodeLong(p+SE_Header)>>8);
			if (Sec>=SPT) {
				PrintError(ReadID, Unit, Track, WrongSecNumber);
				Result = 1;
			}
			else {
				if (CheckCheckSums(p, 1)) {
					PrintError(ReadID, Unit, Track, BadChecksum);
					Result = 1;
				}
				DecodeSector(p+SE_Data, &EBN);
				if (ReallyStore) {
					if (h) {
						if (Seek(h, Track*TRACKSIZE+Sec*TD_SECTOR, OFFSET_BEGINNING)<0) BufferIOError = TRUE;
						if (Write(h, p+SE_Data+TD_SECTOR, TD_SECTOR)<TD_SECTOR) BufferIOError = TRUE;
					}
					else if (VDiskOpen) {
						if (Date && Track == 80 && Sec == 0) UpdateRootBlock(p+SE_Data+TD_SECTOR, TRUE, FALSE);
						VIOStdReq->io_Command = CMD_WRITE;
						VIOStdReq->io_Length = TD_SECTOR;
						VIOStdReq->io_Offset = Track*TRACKSIZE+Sec*TD_SECTOR;
						VIOStdReq->io_Data = (APTR)(p+SE_Data+TD_SECTOR);
						if (DoIO(VIOStdReq)) BufferIOError = TRUE;
					}
					else if (Compression) {
						BlockPointers[(Track/2)%2][(Track%2)*SPT+Sec] = p+SE_Data+TD_SECTOR;
					}
					else CopyMem(p+SE_Data+TD_SECTOR, RAMBuffer[Track/2]+(Track%2)*TRACKSIZE+TD_SECTOR*Sec, TD_SECTOR);
				}
				if (Track == 0 && Sec == 0) DosMark = *((ULONG *)(p+SE_Data+TD_SECTOR));
				if (Track == 80 && Sec == 0) {
					for(Sec=0; Sec<30; Sec++) Name[Sec] = p[SE_Data+TD_SECTOR+433+Sec];
					if ((Sec = (unsigned char)(p[SE_Data+TD_SECTOR+432]))>30) strcpy(Name, ErrorName);
					else Name[Sec] = 0;
				}
			}
			p += BLOCKSIZE;
		}
	else {
		PrintError(ReadID, Unit, Track, WrongNextNumber);
		Result = 1;
	}

	if (BufNumber && ReallyStore) {
		if (Compression) Check[Track] = CheckBlocks(&BlockPointers[(Track/2)%2][SPT*(Track%2)], SPT);
		else Check[Track] = CheckTrack(RAMBuffer[Track/2]+TRACKSIZE*(Track%2), SPT*TD_SECTOR);
	}
	return(Result);
}


/* We take a track from the RAM buffer and we create a MFM one */

signed char RetrieveTrack(unsigned short Track, unsigned char *Buffer) {

	register unsigned char i;
	register unsigned char *p;

	if (Track>159) return(1);

	if (h) {
		if (Track == 0) if (Seek(h, 0, OFFSET_BEGINNING)<0) BufferIOError = TRUE;
	}
	else if (BufNumber) {
		if (!Compression && Check[Track] != CheckTrack(RAMBuffer[Track/2]+TRACKSIZE*(Track%2), SPT*TD_SECTOR)) return(1);
		else if (Compression && Check[Track] != CheckBlocks(&BlockPointers[(Track/2)%2][SPT*(Track%2)], SPT)) return(1);
	}

	for(i=0; i<SPT; i++) {
		p = Buffer+i*BLOCKSIZE;
		SetBufferStart(p);
		EncodeLong(0xFF000000UL | (ULONG)Track<<16 | i<<8 | SPT-i, p+SE_Header);
		setmem(p+SE_Label, TD_LABELSIZE*2, 0xAA);
		if (h) {
			if (Read(h, p+SE_Data+TD_SECTOR, TD_SECTOR)<TD_SECTOR) BufferIOError = TRUE;
		}
		else if (VDiskOpen) {
			VIOStdReq->io_Command = CMD_READ;
			VIOStdReq->io_Length = TD_SECTOR;
			VIOStdReq->io_Offset = Track*TRACKSIZE+i*TD_SECTOR;
			VIOStdReq->io_Data = (APTR)(p+SE_Data+TD_SECTOR);
			if (DoIO(VIOStdReq)) BufferIOError = TRUE;
		}
		else if (!Compression) CopyMemQuick(RAMBuffer[Track/2]+(Track%2)*TRACKSIZE+TD_SECTOR*i, p+SE_Data+TD_SECTOR, TD_SECTOR);
		if (Track == 0 && i == 0 && (h || VDiskOpen)) DosMark = *((ULONG *)(p+SE_Data+TD_SECTOR));
		EncodeSector(p+SE_Data, &EBN);
		EncodeLong(CheckSum(p+SE_Header, (TD_LABELSIZE+4)/2), p+SE_SumL);
		EncodeLong(CheckSum(p+SE_Data, (TD_SECTOR)/2), p+SE_SumD);
	}
	MFMTrack(Buffer, &EBN);
	DMACheck[Track%8] = CheckTrack(Buffer, SPT*BLOCKSIZE);
	SetBufferBorders(Buffer);
	return(0);
}

/* We create a brand new track */

void CreateTrack(unsigned short Track, unsigned char *Buffer, unsigned char UseFFS) {

	register unsigned char i;
	register char *p;

	if (Track>159) return;

	for(i=0; i<SPT; i++) {
		p = Buffer+i*BLOCKSIZE;
		SetBufferStart(p);
		EncodeLong(0xFF000000UL | (ULONG)Track<<16 | i<<8 | SPT-i, p+SE_Header);
		setmem(p+SE_Label, TD_LABELSIZE*2, 0xAA);
		MakeFormatData(Track*SPT+i, p+SE_Data+TD_SECTOR, UseFFS);
		EncodeSector(p+SE_Data, &EBN);
		EncodeLong(CheckSum(p+SE_Header, (TD_LABELSIZE+4)/2), p+SE_SumL);
		EncodeLong(CheckSum(p+SE_Data, TD_SECTOR/2), p+SE_SumD);
	}
	MFMTrack(Buffer, &EBN);
	DMACheck[Track%8] = CheckTrack(Buffer, SPT*BLOCKSIZE);
	SetBufferBorders(Buffer);
}


/* We take an MFM track just read and recode it for writing. */

signed char RecodeTrack(unsigned char *Buffer, unsigned char Unit, unsigned short Track) {

	unsigned char SecNum;
	signed char Result = 0;

	SetBufferStart(Buffer);
	SecNum = SPT-(unsigned char)(DecodeLong(Buffer+SE_Header));
	if ((unsigned char)(DecodeLong(Buffer+SE_Header)>>16) != Track) {
		PrintError(ReadID, Unit, Track, WrongTrackNumber);
		Result = 1;
	}

	if (SecNum>=SPT) {
		PrintError(ReadID, Unit, Track, WrongNextNumber);
		Result = 1;
	}
	else if (SecNum && ArrangeTrack(Buffer+(SPT-SecNum)*BLOCKSIZE, GAPSIZE, SecNum, &EBN)) {
		PrintError(ReadID, Unit, Track, NoFirstSector);
		Result  = 1;
	}
	else if (CheckCheckSums(Buffer, SPT)) {
		PrintError(ReadID, Unit, Track, BadChecksum);
		Result = 1;
	}

	RenumberTrack(Buffer);
	for(SecNum=0; SecNum<SPT; SecNum++) *((ULONG *)(Buffer+SecNum*BLOCKSIZE)) = MFMZeroes;
	MFMTrack(Buffer, &EBN);
	DMACheck[Track%8] = CheckTrack(Buffer, SPT*BLOCKSIZE);
	SetBufferBorders(Buffer);
	return(Result);
}


/* We verify a track. We return 1 on error. */

signed char VerifyTrack(unsigned char *Source, unsigned char *Verify, unsigned char Unit, unsigned short Track) {

	char VStart, *AfterGap;

	Verify += FindOffset(Verify);
	VStart = (char)(DecodeLong(Verify+SE_Header));
	if (VStart<1 || VStart>SPT) {
		PrintError(VerifyID, Unit, Track, WrongNextNumber);
		return(1);
	}
	*((ULONG *)Verify) = *((ULONG *)(Source+(SPT-VStart)*BLOCKSIZE));
	*((ULONG *)Verify+1) = *((ULONG *)(Source+(SPT-VStart)*BLOCKSIZE)+1);

	if (VStart != SPT) {
		if ((AfterGap = FindSync(Verify+VStart*BLOCKSIZE, GAPSIZE)) == NULL) {
			PrintError(VerifyID, Unit, Track, NoFirstSector);
			return(1);
		}
		else {
			AfterGap += FindOffset(AfterGap);
			*((ULONG *)AfterGap) = *((ULONG *)Source);
			*((ULONG *)AfterGap+1) = *((ULONG *)Source+1);
		}
	}

	if (CompareTracks(Source, AfterGap, Verify, SPT-VStart, &EBN)) {
		PrintError(VerifyID, Unit, Track, VerifyError);
		return(1);
	}

	return(0);
}

/* We update the rootblock of an MFM track */

void UpdateMFMRootBlock(unsigned char *Buffer, unsigned char Date, unsigned char IncName) {

	register unsigned char FirstSector, *RootBlock = Buffer;

	FirstSector = (unsigned char)(DecodeLong(Buffer+SE_Header)>>8);
	if (FirstSector>=SPT) return;
	if (FirstSector != 0) RootBlock += (SPT-FirstSector)*BLOCKSIZE;
	DecodeSector(RootBlock+SE_Data, &EBN);
	UpdateRootBlock(RootBlock+SE_Data+TD_SECTOR, Date, IncName);
	EncodeSector(RootBlock+SE_Data, &EBN);
	EncodeLong(CheckSum(RootBlock+SE_Data, TD_SECTOR/2), RootBlock+SE_SumD);
	MFMTrack(Buffer, &EBN);
	DMACheck[80%8] = CheckTrack(Buffer, SPT*BLOCKSIZE);
}


/* From here onwards we aren't using the blitter, so we can call from anywhere. */

/* Standard or MFM or buffer DOS block recognition */

static unsigned char IsDosMark(ULONG Word) {
	if (Word == (((ULONG)('D') << 24) | ((ULONG)('O') << 16) | ((ULONG)('S') << 8) | '\0') ||
		 Word == (((ULONG)('D') << 24) | ((ULONG)('O') << 16) | ((ULONG)('S') << 8) | '\1'))
		return(1);
	else return(0);
}


unsigned char IsDosMFM(unsigned char *Buffer) {

	register unsigned char FirstSector;

	FirstSector = (unsigned char)(DecodeLong(Buffer+SE_Header)>>8);
	if (FirstSector>=SPT) {
/*		printf("Wrong SecNum in IsDos\n");*/
		return(0);
	}
	if (FirstSector != 0) Buffer += (SPT-FirstSector)*BLOCKSIZE;
	return(IsDosMark((*((ULONG *)(Buffer+SE_Data)) & MFMMask)<<1 | (*((ULONG *)(Buffer+SE_Data+TD_SECTOR)) & MFMMask)));
}

unsigned char IsDosBuffered(void) {
	return(IsDosMark(DosMark));
}


/* MFM or buffered disk name getting */

unsigned char *GetNameMFM(unsigned char *Buffer) {

	register unsigned char i;

	i = (char)(DecodeLong(Buffer+SE_Header)>>8);
	if (i>=SPT) strcpy(Name, ErrorName);
	else {
		if (i != 0) Buffer += (SPT-i)*BLOCKSIZE;
		Buffer += SE_Data+432;
		for(i=1; i<31; i++) Name[i-1] = (char)(((((ULONG *)Buffer)[i/4] & MFMMask)<<1 | (((ULONG *)(Buffer+TD_SECTOR))[i/4] & MFMMask))>>(24-(i%4)*8));
		i = (unsigned char)((*Buffer & 0x55)<<1 | (*(Buffer+TD_SECTOR) & 0x55));
		if (i>30) strcpy(Name, ErrorName);
		else Name[i] = 0;
	}
	return(Name);
}

unsigned char *GetNameBuffered(void) {
	return(Name);
}
