#include "SD.h"
#include <rexx/errors.h>

#define TEMPLATE			"PUBSCREEN/K,NOGUI/S,LOWMEM/S\n"
#define NOTRECOGNIZED	"Option not recognized\n"

static unsigned char ARexxConsoleName[] = "CON:0/2/512/128/SuperDuper ARexx Console/AUTO",
							NoHostMsg[] = "ARexx server not active.",
							ARexxErrorMsg[] = "ARexx error: ",
							NotValidMsg[] = "Unknown";

static char RexxCommand[][4] = {	"bu", "hd", "vdb",
											"ve", "da", "inc", "ff", "ta", "au",
											"st", "go", "rea", "ch", "fo", "ki", "res",
											"la", "vdu", "vdn", "ret", "sc", "ec", "qu",
											"so", "de", "req", "vol", "pri", "rx", "no"
											"\0" };

/* This are the opcodes returned by ARexx commands. If they correspond to
	a gadget, it's usually the same letter. For the other ones, we use number codes. */

static unsigned char RexxOp[] = {	'B', 'H', 'U',
												'V', 'D', 'I', '.', 'T', 'A',
												'S', 'G', 'R', 'C', 'F', 'K', 'E',
												1, 2, 3, 4, 5, 6, 'Q',
												7, 8, 9, 10, 11, 12, 13
											};

static unsigned char Switch[2][4] = { "on", "off" };

static BPTR ARexxConsole;
static struct RexxMsg *StartupMsg;

/* This routine launches an ARexx macro selected by the user */


unsigned char StartARexxMacro(unsigned char *p, unsigned char IsStartup) {

	struct RexxMsg *RexxMsg;
	struct MsgPort *HostPort;

	if (p) {
		if (RexxMsg = CreateRexxMsg(ARexxPort, &DefaultDir[8], ARexxPortName)) {
			if (!NewOS) ARexxConsoleName[40] = 0;
			if (stdin && stdout) RexxMsg->rm_Action = RXCOMM;
			else if (ARexxConsole || (ARexxConsole = Open(ARexxConsoleName, MODE_NEWFILE))) {
				RexxMsg->rm_Action = RXCOMM | MASK(RXFB_NOIO);
				RexxMsg->rm_Stdin = RexxMsg->rm_Stdout = ARexxConsole;
			}
			else goto Failure;
			RexxMsg->rm_Args[0] = p;
			Forbid();
			if (HostPort = FindPort("REXX")) PutMsg(HostPort, (void *)RexxMsg);
			Permit();
			if (HostPort) {
				if (IsStartup) StartupMsg = RexxMsg;
				return(1);
			}
			if (!IsStartup) Acknowledge(NoHostMsg);
		}
	}
	else return(0);
Failure:
	DeleteArgstring(p);
	if (RexxMsg) DeleteRexxMsg(RexxMsg);
	return(0);
}

void CloseARexxConsole(void) {
	if (ARexxConsole) {
		Close(ARexxConsole);
		ARexxConsole = NULL;
	}
}

/* After an arexx macro... */

void AfterARexxMacro(struct RexxMsg *RexxMsg) {

	struct NexxStr *p;

	if (RexxMsg->rm_Result1 && (StartupMsg != RexxMsg || RexxMsg->rm_Result2 != ERR10_001)) {
		ObtainSemaphore(&Semaphore);
		strcpy(s, ARexxErrorMsg);
		if (ErrorMsg(RexxMsg->rm_Result2, (void *)&p)) strcat(s, p->ns_Buff);
		else strcat(s, NotValidMsg);
		Acknowledge(s);
		ReleaseSemaphore(&Semaphore);
	}
	DeleteArgstring(RexxMsg->rm_Args[0]);
	DeleteRexxMsg(RexxMsg);
	if (RexxMsg == StartupMsg) StartupMsg = NULL;
}


/* This is Tom's wonderful function for matching the *c command against the *m pattern */

static char cmdcmp(char *c, char *m) {
   while (*c && ((*c == *m) || (*c == *m + 32 && ('a' <= *c && *c <= 'z')))) {
      c++ ;
      m++ ;
   }
   return(*c) ;
}

/*
*
* ParseARexx()
*
* takes an ARexxMessage pointer and the address of a pointer which will
* be set to the argument.
*
*/

unsigned char ParseARexx(struct RexxMsg *ARexxMessage, unsigned char **Arg) {

	register unsigned short int i=0;
	register unsigned char *p;

	do {
		if (!cmdcmp(RexxCommand[i], p = ARexxMessage->rm_Args[0])) {
			while(*(p++)>' ');
			if (!*(p-1)) p--;
			*Arg = p;
			return(RexxOp[i]);
		}
	} while(RexxCommand[++i][0]);

	return(0);
}

unsigned char SwitchOn(unsigned char *p) {
	if (!cmdcmp(Switch[0], p)) return(1);
	if (!cmdcmp(Switch[1], p)) return(0);
	return(2);
}

char *SkipSpace(char *s) {
	while(*s == ' ' || *s == '\n' || *s == '\t') s++;
	return(s);
}

char *SkipNotSpace(char *s) {
	while(*s && *s != ' ' && *s != '\t' && *s != '\n') s++;
	return(s);
}

unsigned char ParseCommandLine(unsigned char *l) {

	unsigned char *p, NextIsScreen = 0, NotRec = 0;
	int len;

	if (!*l) return(0);
	if (*l == '\"') {
		l++;
		while(*l && *l != '\"') l++;
		if (*l) l++;
	}
	else l = SkipNotSpace(l);
	l = SkipSpace(l);
	if (!*l) return(0);
	if (*l == '?') {
		Write(stdout, TEMPLATE, strlen(TEMPLATE));
		return(1);
	}

	len = strlen(l);

	for(;;) {
		p = l;
		l = SkipNotSpace(l);

		*(l++) = 0;
		if (NextIsScreen) {
			PubScreen = p;
			NextIsScreen = FALSE;
		}
		else if (!cmdcmp("lowmem", p)) ARexx = Sound = FALSE;
		else if (!cmdcmp("nogui", p)) GUI = FALSE;
		else if (!cmdcmp("pubscreen", p)) NextIsScreen = TRUE;
		else {
			NotRec = TRUE;
			break;
		}

		l = SkipSpace(l);

		if ( (len -= (l-p)) <= 0) break;
	}

/*	Write(stdout, PubScreen, strlen(PubScreen));*/

	if (NextIsScreen || NotRec || ARexx == FALSE && GUI == FALSE) {
		Write(stdout, NOTRECOGNIZED, strlen(NOTRECOGNIZED));
		return(1);
	}

	return(0);
}
