#include "SD.h"

#define ESC_CODE 0x1B

#define GADGETMASK	(0x3FBFFF00UL)
#define COPYMASK 		(0x0BC3E000UL)
#define READMASK		(0x0BC3F800UL)
#define CHECKMASK		(0x0BC3F800UL)
#define FORMATMASK	(0x0AC38000UL)

void	FakeSegList(void);
void __saveds CopyTask(void);

long	stdin, stdout;	/* for debug I/O & ARexx */
char	NewOS, 			/* Are we under 2.0? */
		GUI = 1,			/* Do we have a GUI? */
		ARexx = 1,		/* This is set at startup to "do we want ARexx?".
								In case we want it, the main allocation routines
								sets it to "can the ARexx gadget work?" */
		Sound = 1,		/* This remind us "do we want sound?" */
		Compression,	/* Are we using compression? */
		MungedBuffer,	/* Is the current buffer munged ? */
		BufferIOError,
		Requesters = 1,/* Do we want system requesters to appear? */
		PrintErrors,	/* Do we want printed output of disk errors? */
		NoClick;			/* Mask of the units to force NOCLICK */

unsigned short Errors[4],			/* Number of errors for each unit */
					Retries[4],			/* Number of retries for each unit */
					Copies;				/* Number of buffered copies up to now */

struct RxsLib *RexxSysBase;

struct ExtendedBlitNode EBN;		/* A special blitnode for all our blitter operations */

unsigned char *ReadBuffer[3], *VerifyBuffer[2]; /* "real" read and verify buffers; they're
																	normally accessed through VB[] and B[],
																	so that offsetting for SYNCs and gaps
																	is included automatically. */

struct DiskBlockServerData volatile DiskBlockServerData; /* This is where we pass data to the interrupt
																				DSKBLK server */

static unsigned char NoBufferMsg[] = "Sorry, I need a buffer.",
							NoFileMsg[] = "Can't allocate the file.",
							NoMemoryMsg[] = "Not enough memory.",
							NoDeviceMsg[] = "Can't open the requested device.",
							NoDestMsg[] = "Sorry, I need a destination.",
							NoSourceMsg[] = "Sorry, I need a source.",
							NoDiskPresentMsg[] = "No disk in unit x.",
							WriteProtectedMsg[] = "Disk in unit x is write protected.",
							ChecksumErrorMsg[] = "Checksum error: buffer munged.",
							InvalidBufferMsg[] = "The buffer is not valid.",
							ErrorsMsg[] = "There were errors.",
							WrongTrackNumbersMsg[] = "Track range not valid.",
							ProtectSourceMsg[] = "Better write-protect your sources.",
							RangeMismatchMsg[] = "No buffer data in this range.",
							BufferMungedMsg[] = "A track buffer has been munged.",
							BufferIOErrorMsg[] = "Buffer I/O error.",
							StartupName[] = "Startup.supdup",
							GfxName[] = "graphics.library",
							UNKNOWN[] = "<UNKNOWN>",
							NOTDOS[] = "<NDOS>",
							CopyTaskName[] = "SD_Copier";

static struct timeval a, b; /* for timing */

struct Resource *DiskBase;
struct DiscResourceUnit DRU = {
	{ { NULL, NULL, NT_MESSAGE, 127, "SD_DRU" }, NULL, sizeof(struct DiscResourceUnit)},
	{ { NULL, NULL, NT_INTERRUPT, 10, NULL }, (APTR)&DiskBlockServerData, InterruptCode },
	{ { NULL, NULL, NT_INTERRUPT, 10, NULL }, (APTR)&SyncIntData, SyncCode },
	{ { NULL, NULL, NT_INTERRUPT, 0, NULL }, NULL, NULL }
};

static unsigned char CIA,					/* For holding CIA registers values */
							Dest,					/* Current destination mask (from gadgets) */
							ActualDest,			/* "Real" destination mask in the current operation */
							ActualSource,		/* "Real" source number for the current copy */
							Verify = 1,			/* Is verify on? */
							Buffer,				/* Are we buffering? */
							Formatting,			/* Are we formatting? */
							Auto,					/* Is auto on? */
							Date = 1,			/* Is date adaption on? */
							UseFFS,				/* Is FFS on? */
							Copying,				/* What are we doing? ('C','R', etc.) */
							GapOpt,				/* Is the gap verify optimization on? */
							RetryNumber = 4,	/* How many retries? */
							DestNumber,			/* How many destination units? */
							DestDrive[4],		/* What are the destination drives? */
							ValidBuffer,		/* Is the current buffer valid? */
							AutoFlag,			/* Should we autostart ? */
							IsDos,				/* Is the current disk a DOS disk? */
							BufferIsDos,		/* Is the current buffer a DOS disk? */
							Unknown,				/* Is the name of the current disk unknown? */
							Modulo,				/* What's the modulo for accessing buffers? */
							DrawColor,			/* In what color are we drawing the bar? */
							PresentDisks,		/* What disks are linked to the system? */
							DisksOut,			/* What disks have been pulled out (at least 1 time)? */
							DisksIn,				/* What disks are in? */
							IncName,				/* Should we increment the format label? */
							MultipleRootBlocks,	/* Do we have multiple root blocks? (e.g., for date adaption) */
							ChecksumError,		/* The RAM buffer is trashed */
							Stopped;				/* Set by the copy task if the copy operation was stopped */

static unsigned char volatile Done;		/* Set this to one when everything is over */

static signed char	Source = -1,	/* The number of the source */
							Offset[3];		/* What offset are we using when accessing a buffer? */

static unsigned short	StartTrack,			/* First track to copy */
								EndTrack = 159,	/* Last Track to copy */
								ST = 0,				/* Selected start track */
								ET = 159,			/* Selected end track */
								BST,					/* Buffer start track */
								BET,					/* Buffer end track */
								FailedGaps,			/* How many gap optimization have failed? */
								Gap[4],				/* How long is the gap? */
								WriteGap,			/* Write gap length */
								PendingMacros;

static unsigned char *VB[2], *B[5];  /*	The buffers, offset in such a way
														you can set SYNCs and inter-sector
														gaps *after* reading in */

/*
 * Masks for EnableGadgets. They specify what button gadgets should be enabled
 * during an operation. The LSB is the df0: source gadget, and so on.
 */

static unsigned long CurrentMask =	GADGETMASK,
							GadgetMask =	GADGETMASK,
							ClickCount,		/* How many time click have passed? */
							Signals;			/* The signals we get in the main loop */

unsigned long	StartMask,
					StopMask,
					CopierMask;						/* Masks for copy task <-> main task communication */

static struct RexxMsg	*PendingRexxMsg,	/* Do we have a Rexx message pending? */
								*RexxMsg,			/* To read the ARexx port */
								*RemRexx;			/* To remember if something was read from the ARexx port */


void FlushVars(void) {}

/* Here we set the error fields of an ARexx message */

static void SetError(struct RexxMsg *RM, unsigned short n) {
	if (RM) {
		RM->rm_Result1 = n;
		RM->rm_Result2 = NULL;
	}
}

/* Here we reset all vars related to automatic operation start */

static void ResetAuto(void) {
	DisksOut = 0;
	AutoFlag = 0;
}

/* Here we stop all motor we are using */

static void StopMotorsInUse(void) {
	MotorOff(ActualDest | MASK(ActualSource));
}

/* Here we start all motors we are using */

static void StartMotorsInUse(void) {
	MotorOn(ActualDest | MASK(ActualSource));
}

/* This function is called after any copy operation. It a) tells (if possible)
	the user that the operation is complete, b) enables the gadgets,
	c) if there is no ARexx message pending displays an error requester if it's
	the case, otherwise sets the return codes. Copying is set to 0, and a
	ResetAuto() is performed. This function has to be called from the main task. */

static void AfterCopy(void) {

	register unsigned char i;

	if (!Stopped && !MungedBuffer && !ChecksumError && !BufferIOError) {
		if (Buffer && ((!Formatting && Copying == 'G') || Copying == 'R')) Say("Pass complete.");
		else Say("Operation complete.");
	}

	if (ChecksumError) Acknowledge(ChecksumErrorMsg);
	if (MungedBuffer) Acknowledge(BufferMungedMsg);
	if (BufferIOError) Acknowledge(BufferIOErrorMsg);
	if (Stopped && Copying == 'R') {
		ValidBuffer = FALSE;
		WipeProgress(StartTrack/2);
		BET = BST-1;
	}

	EnableGadgets(CurrentMask = GadgetMask, Formatting && Copying == 'G' ? 'F' : Copying);
	for(i=0; i<4; i++) {
		if (((MASK(i) & ActualDest) || i == ActualSource) && Errors[i]) {
			Acknowledge(ErrorsMsg);
			break;
		}
	}
	if (PendingRexxMsg) {
		if (MungedBuffer || ChecksumError || BufferIOError) SetError(PendingRexxMsg, 20);
		else if (i<4) SetError(PendingRexxMsg, (Compression && BET != ET) ? 9 : 8);
		else if (Compression && BET != ET) SetError(PendingRexxMsg, 2);
		ReplyMsg((void *)PendingRexxMsg);
		PendingRexxMsg = NULL;
	}
	BufferIOError = MungedBuffer = ChecksumError = Stopped = FALSE;
	ResetAuto();
	Copying = 0;
	UpdateProgress(FALSE);
}

/*	Here we stop any operation. If we weren't copying, and the buffer is on,
	and it's not valid OR it's partial, we wipe the progress bar and reset
	the pass counter. You can call this function from everywhere. */

static void Stop(void) {

	if (Copying) Signal(Copier, StopMask);
	else {
		if (!ValidBuffer || BET<ET || BST>ST) {
			WipeProgress(0);
			ValidBuffer = 0;
			BET = ET;
		}
	}
}



/* Here we disable/enable a gadget */

static void DisableGadgetMask(unsigned long Mask) {
	CurrentMask &= ~Mask;
	GadgetMask &= ~Mask;
	EnableGadgets(CurrentMask, 0);
}

static void EnableGadgetMask(unsigned long Mask) {
	CurrentMask |= Mask;
	GadgetMask |= Mask;
	EnableGadgets(CurrentMask, 0);
}

/* Here we try to get or free a buffer, depending on Sel. */

static void GetBuffer(unsigned char Sel, unsigned char WhichBuffer, unsigned char WhichGadget, unsigned char *ErrMsg) {

	if (Sel) {
		if (Buffer == WhichBuffer) return;
		else if (Buffer) {
			DeselectBufferGadgets(WhichGadget);
			FreeAllBuffers();
			Buffer = NO_BUFFER;
		}
		if ((WhichBuffer == RAM_BUFFER && !AllocBuffers()) ||
			(WhichBuffer == HD_BUFFER && !AllocHDBuffer()) ||
			(WhichBuffer == VDISK_BUFFER && !AllocVDisk())) {
			Acknowledge(ErrMsg);
			SelectGadget(GAddr(WhichGadget), FALSE);
			SetError(RexxMsg, 5);
		}
		else {
			Buffer = WhichBuffer;
			if (Compression) SetError(RexxMsg, 2);
		}
	}
	else {
		FreeAllBuffers();
		Buffer = NO_BUFFER;
	}
	if (Buffer == VDISK_BUFFER) DisableGadgetMask(MASK(VDNAME) | MASK(VDNUMBER));
	else EnableGadgetMask(MASK(VDNAME) | MASK(VDNUMBER));
	BST = 0;
	BET = 159;
	BufferIsDos = Copies = ValidBuffer = 0;
	WriteCopies();
	ResetAuto();
}

/* Here we draw the dragbar, basing on the track number, and we update
	the copy time. This has to be called from the copy task. */

static void DrawBar(unsigned short Tr) {
	if (GUI && Tr%2) {
		DrawProgress(Tr/2, DrawColor);
		CopyTaskGetSystemTime(&b);
		WriteElapsed(&a, &b);
	}
}

/* Here we verify a track and in case we retry it */

static void VerifyAndRetry(unsigned char Unit, unsigned short Tr, unsigned char *VBuffer) {

	unsigned char i = 0;

	if (VerifyTrack(B[Tr%Modulo]+Offset[Tr%Modulo], VBuffer, Unit, Tr)) {
		if (RetryNumber) {
			MoveAndBackupTrack(Unit, Tr);
			do {
				if (i++ == RetryNumber) break;
				Retries[Unit]++;
				WriteRetriesErrors(Unit, 160);
				StartWrite(MASK(Unit), Tr, B[Tr%Modulo]+Offset[Tr%Modulo]-WriteGap, WRITELENGTH+WriteGap, Unit, VBuffer, TRUE);
				WaitDMA();
			} while(VerifyTrack(B[Tr%Modulo]+Offset[Tr%Modulo], VBuffer, Unit, Tr));
			RestoreTrack(Unit);
		}
		else i = 1;

		if (i == RetryNumber+1) {
			Retries[Unit]-=RetryNumber;
			Errors[Unit]++;
			WriteRetriesErrors(Unit, Tr);
			Beep(6000, TRUE);
		}
	}
}

/* Here we recode a track just read for writing. If something fails, we
	retry it. */

static void RecodeAndRetry(unsigned short Tr) {

	unsigned char i = 0;

	Offset[Tr%Modulo] = FindOffset(B[Tr%Modulo]);
	if (RecodeTrack(B[Tr%Modulo]+Offset[Tr%Modulo], ActualSource, Tr)) {
		if (RetryNumber) {
			MoveAndBackupTrack(ActualSource, Tr);
			do {
				if (i++ == RetryNumber) break;
				Retries[ActualSource]++;
				WriteRetriesErrors(ActualSource, 160);
				StartRead(ActualSource, Tr, B[Tr%Modulo], FALSE);
				WaitDMA();
				Offset[Tr%Modulo] = FindOffset(B[Tr%Modulo]);
			} while(RecodeTrack(B[Tr%Modulo]+Offset[Tr%Modulo], ActualSource, Tr));
			RestoreTrack(ActualSource);
		}
		else i = 1;

		if (i == RetryNumber+1) {
			Retries[ActualSource]-=RetryNumber;
			Errors[ActualSource]++;
			WriteRetriesErrors(ActualSource, Tr);
			Beep(6000, TRUE);
		}
	}
}

/* Here we store a track just read, and in case we retry. */

static void StoreAndRetry(unsigned short Tr) {

	unsigned char i = 0;

	if (StoreTrack(B[Tr%5], ActualSource, Tr, Copying == 'R', IsDos && Date)) {
		if (RetryNumber) {
			MoveAndBackupTrack(ActualSource, Tr);
			do {
				if (i++ == RetryNumber) break;
				Retries[ActualSource]++;
				WriteRetriesErrors(ActualSource, 160);
				StartRead(ActualSource, Tr, B[Tr%5], FALSE);
				WaitDMA();
			} while(StoreTrack(B[Tr%5], ActualSource, Tr, Copying == 'R', IsDos && Date));
			RestoreTrack(ActualSource);
		}
		else i = 1;

		if (i == RetryNumber+1) {
			Beep(6000, TRUE);
			Retries[ActualSource]-=RetryNumber;
			Errors[ActualSource]++;
			WriteRetriesErrors(ActualSource, Tr);
		}
	}
}


/* Here we get a track from the buffer, and if the checksum is wrong we stop
	the copy. */

static void RetrieveAndCheck(unsigned short Tr) {

	if (RetrieveTrack(Tr, B[Tr%Modulo])) ChecksumError = TRUE;
}
/* Here we do the action corresponding to an option gadget */

static void SwitchOp(short Op, unsigned short i) {
	switch(Op) {
		case 'V':	Verify = i;
						break;
		case 'F':	Formatting = i;
						ResetAuto();
						break;
		case 'D':	Date = i;
						break;
		case '.':	UseFFS = i;
						break;
		case 'A':	Auto = i;
						ResetAuto();
						if (i && (PresentDisks & 0xF) == 1) Acknowledge(ProtectSourceMsg);
						break;
		case 'I':	IncName = i;
						break;
		case 'T':	if (i) {
							if (!OpenVoice()) {
								SelectGadget(GAddr(TALK), FALSE);
								SetError(RexxMsg, 5);
							}
						}
						else CloseVoice();
						break;
		case 'B':	GetBuffer(i, RAM_BUFFER, BUFFER, NoMemoryMsg);
						break;
		case 'H':	GetBuffer(i, HD_BUFFER, HDBUF, NoFileMsg);
						break;
		case 'U':	GetBuffer(i, VDISK_BUFFER, VDBUF, NoDeviceMsg);
						break;

	}
}


/* Reset the copy state to an unknown disk */

static void ResetToUnknown(void) {
	IsDos = 0;
	if (!Unknown) AddName(UNKNOWN);
	Unknown = 1;
}

static unsigned char ShiftedKey(unsigned char Op, ULONG Class, struct Gadget *G) {

	unsigned short i;

	for(i=0; i<4; i++)
		if (ShiftedKeyCodes[i] == Op) {
			if (!G) return(1);
			if (Class == RAWKEY) ToggleGadget(G);

			if (G->Flags & SELECTED) {
				if (GrabUnits(PresentDisks)) Dest |= MASK(i);
				else DeselectDiskGadgets();
			}
			else Dest &= ~MASK(i);
			if (!Dest && Source<0) DropUnits(PresentDisks);
			ResetAuto();
			return(1);
		}

	for(i=0; i<4; i++)
		if (UnshiftedKeyCodes[i] == Op) {
			if (!G) return(1);
			if (Class == RAWKEY) ToggleGadget(G);

			if (G->Flags & SELECTED) {
				if (Source == i) return(1);
				if (Source>=0) SelectGadget(GAddr(DF0S)+Source, FALSE);
				if (GrabUnits(PresentDisks)) Source = i;
				else DeselectDiskGadgets();
			}
			else if (i == Source) Source = -1;
			if (!Dest && Source<0) DropUnits(PresentDisks);
			ResetAuto();
			return(1);
		}

	return(0);
}

/* Here we take the action corresponding to an action gadget */

static unsigned char SwitchActions(short Op) {
	unsigned short i;

	if (Copying) return(1);
	ResetAuto();
	if ((i = GetSCyl()*2) != ST) WipeProgress(0);
	ST = i;
	if ((i = GetECyl()*2+1) != ET) WipeProgress(0);
	ET = i;
	if (ET<ST || ET>159) {
		Acknowledge(WrongTrackNumbersMsg);
		return(7);
	}
	if (((Op == 'G' && !Buffer && !Formatting) || Op == 'R' || Op == 'C') && Source<0) {
		Acknowledge(NoSourceMsg);
		return(3);
	}
	if (((Op == 'G' && INDEST(Source) && !Formatting) || Op == 'R') && !Buffer ) {
		Acknowledge(NoBufferMsg);
		return(3);
	}
	if (Op == 'G' && !Dest) {
		Acknowledge(NoDestMsg);
		return(3);
	}
	if (Op == 'G' && (Buffer == RAM_BUFFER) && !Formatting) {
		if (ValidBuffer) {
			if (ST>BET || ET<BST) {
				Acknowledge(RangeMismatchMsg);
				return(7);
			}
		}
		else {
			Acknowledge(InvalidBufferMsg);
			return(4);
		}
	}
	ActualDest = Dest;
	ActualSource = Source;
	if ((Copying = Op) == 'R' || Copying == 'C') ActualDest = MASK(4);
	else if (Buffer || Formatting) ActualSource = 4;

	FailedGaps = 0;
	GapOpt = 1;
	WriteGap = GAPSIZE;

	for(i=0; i<4; i++) {
		Errors[i] = Retries[i] = 0;
		WriteRetriesErrors(i, 160);
		if (INACTUALDEST(i) || i==ActualSource) {
			CIA = GetReallyCIAAPRA(i);
			if (!(CIA & CIAF_DSKCHANGE)) {
				NoDiskPresentMsg[16] = 48+i;
				Acknowledge(NoDiskPresentMsg);
				Copying = 0;
				return(5);
			}
			if (INACTUALDEST(i) && !(CIA & CIAF_DSKPROT)) {
				WriteProtectedMsg[13] = 48+i;
				Acknowledge(WriteProtectedMsg);
				Copying = 0;
				return(6);
			}
			DisksIn |= MASK(i);
		}
	}

	TimerWait(SettleDelay);
	DrawColor = 3;
	Modulo = 3;
	StartTrack = ST;
	EndTrack = ET;
	for(i=0; i<3; i++) B[i] = ReadBuffer[i]+GAPSIZE+8;
	if (Copying == 'R' || Copying == 'C') {
		DrawColor = 1;
		if (Copying == 'C') ResetToUnknown();
		else if (Buffer == VDISK_BUFFER || Buffer == HD_BUFFER) {
			ResetToUnknown();
			BufferIsDos = 0;
		}
		else if (!Compression || BET >= ET) {
			BST = StartTrack;
			BET = EndTrack;
			ResetToUnknown();
			BufferIsDos = 0;
		}
		else {
			BST = StartTrack = max(ST, BET+1);
			BET = EndTrack;
		}
		Modulo = 5;
	}
	else if (!Buffer && !Formatting) ResetToUnknown();
	else {
		if (Formatting) {
			IsDos = 1;
			DrawColor = 2;
		}
		else {
			IsDos = BufferIsDos;
			StartTrack = max(BST, ST);
			EndTrack = min(ET, BET);
		}
		for(i=0; i<3; i++) {
			B[i] = ReadBuffer[i]+GAPSIZE;
			Offset[i] = 0;
		}
	}

	DestNumber = 0;
	for(i=0; i<4; i++)
		if (INACTUALDEST(i)) {
			DestDrive[DestNumber++] = i;
		}
	if ((Copying == 'R' || (Copying == 'G' && !Formatting)) && Compression) SetUpCompression(Copying == 'R');

	GetSystemTime(&a);
	StartMotorsInUse();
	SetUpDisksA();
	if (Copying == 'R') {
		SelectGadget(GAddr(FORMAT), FALSE);
		Formatting = 0;
		EnableGadgets(CurrentMask = GadgetMask & READMASK, 'R');
	}
	else if (Copying == 'C') EnableGadgets(CurrentMask = GadgetMask & CHECKMASK, 'C');
	else if (Formatting) EnableGadgets(CurrentMask = GadgetMask & FORMATMASK, 'F');
	else EnableGadgets(CurrentMask = GadgetMask & COPYMASK, 'G');
	WipeProgress(StartTrack/2);
	SetUpDisksB(ActualDest | MASK(ActualSource), StartTrack);

	RetryNumber = GetRetryNumber();
	MultipleRootBlocks = Date || (IncName && Formatting);

	Signal(Copier, StartMask);

	return(0);
}

void __saveds CopyTask(void) {

	register unsigned int i;
	register unsigned short Track;
	BPTR CDLock;

	DiskBlockServerData.DBSD_ISD.ISD_Task = SoftIntData.SID_ISD.ISD_Task = EBN.EBN_Task = Copier = FindTask(NULL);
	if (!(Done = !(CopyTaskAllocStuff() && CopyTaskOpenTimer())) && Sound) SetUpAudio();
	CDLock = CurrentDir(DupLock(((struct Process *)Me)->pr_CurrentDir));
	Signal(Me, CopierMask);

	FOREVER {
		Wait(StartMask);
		if (Done) break;
		SetSignal(0, StopMask);
		FlushVars();

		Track = StartTrack;
		if (Copying == 'R' || (Copying == 'G' && !Formatting)) {
			if (Buffer == VDISK_BUFFER) BufferIOError = !CopyTaskOpenVDisk();
			if (Buffer == HD_BUFFER) BufferIOError = !CopyTaskOpenHDBuffer();
		}

		if (!BufferIOError) {
			if (Copying == 'R' || Copying == 'C') {
				StartRead(ActualSource, Track, B[Track%5], FALSE);
			}
			else if (!Buffer && !Formatting) {
				StartRead(ActualSource, Track, B[Track%3], FALSE);
				MoveHead(ActualSource, Track+1);
			}
			else if (Formatting) {
				CreateTrack(Track, B[Track%3], UseFFS);
			}
			else {
				if (Compression) {
					if (BST<ST)
						for(i=BST; i<ST; i+=2) {
							StartDecompression(i, B[i%3], B[(i+1)%3]);
							WaitInterDecompression();
							WaitCompression(FALSE);
						}
					StartDecompression(Track, B[Track%3], B[(Track+1)%3]);
					WaitInterDecompression();
				}
				RetrieveAndCheck(Track);
			}
			Track++;
		}

		FOREVER {
			if (MungedBuffer || ChecksumError || BufferIOError || (SetSignal(0,0) & StopMask)) {
				Beep(2400, FALSE);
				Stopped = TRUE;
				break;
			}
			if (Track > 1+EndTrack) break;
			if (Track == EndTrack-8 && Track > StartTrack+6) Beep(3000, FALSE);
			if (Copying == 'R' || Copying == 'C') {
				if (Track<=EndTrack) {
					MoveHead(ActualSource, Track);
					StartRead(ActualSource, Track, B[Track%5], FALSE);
				}
				else WaitDMA();
				DrawBar(Track);
				StoreAndRetry(Track-1);
				if (Copying == 'R') {
					if (Compression && !(Track%2)) {
						if (WaitCompression(FALSE)) {
							EndTrack = Track-5;
							WipeProgress((EndTrack+1)/2);
							continue;
						}
						StartCompression(Track-2);
					}
				}
				if (Track == 81 && IsDos) {
					ChangeName(GetNameBuffered());
					Unknown = 0;
				}
				if (Track == 1 && !(BufferIsDos = IsDos = IsDosBuffered())) {
					ChangeName(NOTDOS);
					Unknown = 0;
				}
			}
			else if (!Buffer && !Formatting) {
				if (Verify)	{
					if (Track<=EndTrack) {
						StartRead(ActualSource, Track, B[Track%3], FALSE);
						DrawBar(Track+1-(Track == StartTrack+1));
					}
					if (Track == StartTrack+1) {
						RecodeAndRetry(StartTrack);
						if (StartTrack == 0 && !(IsDos = IsDosMFM(B[0]+Offset[0]))) {
							ChangeName(NOTDOS);
							Unknown = 0;
						}
					}
					else {
						VerifyAndRetry(DestDrive[DestNumber-1], Track-2, VB[1-DestNumber%2]);
						if (Track == StartTrack+2) {
							Gap[DestDrive[DestNumber-1]] = GapLength(VB[1-DestNumber%2]);
							WriteGap = 0;
							for(i=0; i<DestNumber; i++) if (Gap[DestDrive[i]]>WriteGap) WriteGap = Gap[DestDrive[i]];
						}
					}
					if (Track == 81 && Date && IsDos) UpdateMFMRootBlock(B[80%3]+Offset[80%3], TRUE, FALSE);
					if (Track == 81 && IsDos) {
						ChangeName(GetNameMFM(B[80%3]+Offset[80%3]));
						Unknown = 0;
					}
					if (Track<EndTrack) MoveHead(ActualSource, Track+1);
					if (Track != 81 || !Date || !IsDos) StartWrite(ActualDest, Track-1, B[(Track-1)%3]+Offset[(Track-1)%3]-WriteGap, WRITELENGTH+WriteGap, DestDrive[0], VB[0], Track == StartTrack+1 || !GapOpt);
					else {
						StartWrite(MASK(DestDrive[0]), 80, B[80%3]+Offset[80%3]-WriteGap, WRITELENGTH+WriteGap, DestDrive[0], VB[0], StartTrack == 80 || !GapOpt);
					}
					if (Track<=EndTrack) RecodeAndRetry(Track);
					WaitDMA();
					if (GapOpt && (char)(DecodeLong(VB[0]+FindOffset(VB[0])+SE_Header)) != SPT) {
						StartRead(DestDrive[0], Track-1, VB[0], Track == StartTrack+1);
						FailedGaps++;
						if (FailedGaps>(Track-StartTrack)/10+1) GapOpt = 0;
						WaitDMA();
					}
					if (Track<=EndTrack) MoveHead(DestDrive[0], Track);
					for (i=1; i<DestNumber; i++) {
						if (Track != 81 || !Date || !IsDos) StartRead(DestDrive[i], Track-1, VB[i%2], Track == StartTrack+1);
						VerifyAndRetry(DestDrive[i-1], Track-1, VB[1-i%2]);
						if (Track == StartTrack+1) Gap[DestDrive[i-1]] = GapLength(VB[1-i%2]);
						if (Track == 81 && Date && IsDos) {
							UpdateMFMRootBlock(B[80%3]+Offset[80%3], TRUE, FALSE);
							StartWrite(MASK(DestDrive[i]), Track-1, B[80%3]+Offset[80%3]-WriteGap, WRITELENGTH+WriteGap, DestDrive[i], VB[i%2], TRUE);
						}
						WaitDMA();
						if (Track<=EndTrack) MoveHead(DestDrive[i], Track);
					}
					if (Track>EndTrack) VerifyAndRetry(DestDrive[DestNumber-1], EndTrack, VB[1-DestNumber%2]);
				}
				else {
					if (Track<=EndTrack) {
						StartRead(ActualSource, Track, B[Track%3], FALSE);
						DrawBar(Track+1-(Track == StartTrack+1));
					}
					if (Track == StartTrack+1) {
						RecodeAndRetry(StartTrack);
						if (StartTrack == 0 && !(IsDos = IsDosMFM(B[0]+Offset[0]))) {
							ChangeName(NOTDOS);
							Unknown = 0;
						}
					}
					if (Track == 81 && IsDos) {
						ChangeName(GetNameMFM(B[80%3]+Offset[80%3]));
						Unknown = 0;
					}
					if (Track<EndTrack) MoveHead(ActualSource, Track+1);
					if (Track != 81 || !Date || !IsDos) StartWrite(ActualDest, Track-1, B[(Track-1)%3]+Offset[(Track-1)%3]-WriteGap, WRITELENGTH+WriteGap, DestDrive[0], (Track == StartTrack+1) ? VB[0] : NULL, TRUE);
					else for(i=0; i<DestNumber; i++) {
						WaitDMA();
						UpdateMFMRootBlock(B[(Track-1)%3]+Offset[(Track-1)%3], TRUE, FALSE);
						StartWrite(MASK(DestDrive[i]), Track-1, B[(Track-1)%3]+Offset[(Track-1)%3]-WriteGap, WRITELENGTH+WriteGap, DestDrive[0], (i == 0 && Track == StartTrack+1) ? VB[0] : NULL, TRUE);
					}
					if (Track == StartTrack+1) {
						for(i=1; i<DestNumber; i++) {
							StartRead(DestDrive[i], Track-1, VB[i%2], TRUE);
							Gap[DestDrive[i-1]] = GapLength(VB[(i-1)%2]);
						}
						WaitDMA();
						Gap[DestDrive[i-1]] = GapLength(VB[(i-1)%2]);
						WriteGap = 0;
						for(i=0; i<DestNumber; i++) if (Gap[DestDrive[i]]>WriteGap) WriteGap = Gap[DestDrive[i]];
					}
					if (Track<=EndTrack) {
						RecodeAndRetry(Track);
						MoveHeads(ActualDest, Track);
					}
				}
			}
			else {
				if (Verify) {
					if (Track != 81 || !MultipleRootBlocks || !IsDos) StartWrite(ActualDest, Track-1, B[(Track-1)%3]-WriteGap, WRITELENGTH+WriteGap, DestDrive[0], VB[DestNumber%2 ? Track%2 : 0], Track == StartTrack+1 || !GapOpt);
					else {
						UpdateMFMRootBlock(B[80%3], Date, IncName);
						StartWrite(MASK(DestDrive[0]), 80, B[80%3]-WriteGap, WRITELENGTH+WriteGap, DestDrive[0], VB[DestNumber%2], StartTrack == 80 || !GapOpt);
					}
					DrawBar(Track);
					if (Track>StartTrack+1) {
						VerifyAndRetry(DestDrive[DestNumber-1], Track-2, VB[DestNumber%2 ? (Track+1)%2 : 1]);
						if (Track == StartTrack+2) {
							Gap[DestDrive[DestNumber-1]] = GapLength(VB[DestNumber%2 ? (Track+1)%2 : 1]);
							WriteGap = 0;
							for(i=0; i<DestNumber; i++) if (Gap[DestDrive[i]]>WriteGap) WriteGap = Gap[DestDrive[i]];
						}
					}
					if (Track<=EndTrack) {
						if (Formatting) CreateTrack(Track, B[Track%3], UseFFS);
						else {
							if (Compression) {
								if (Track%2) {
									WaitCompression(Track == EndTrack);
									if (Track != EndTrack) StartDecompression(Track+1, B[(Track+1)%3], B[(Track+2)%3]);
								}
								else WaitInterDecompression();
							}
							RetrieveAndCheck(Track);
						}
					}
					WaitDMA();
					if (GapOpt && (char)(DecodeLong(VB[DestNumber%2 ? Track%2 : 0]+FindOffset(VB[DestNumber%2 ? Track%2 : 0])+SE_Header)) != SPT) {
						StartRead(DestDrive[0], Track-1, VB[DestNumber%2 ? Track%2 : 0], Track == StartTrack+1);
						FailedGaps++;
						if (FailedGaps>(Track-StartTrack)/10+1) GapOpt = 0;
						WaitDMA();
					}
					if (Track<=EndTrack) MoveHead(DestDrive[0], Track);
					for (i=1; i<DestNumber; i++) {
						if (Track != 81 || !MultipleRootBlocks || !IsDos) StartRead(DestDrive[i], Track-1, VB[DestNumber%2 ? (i+Track)%2 : i%2], Track == StartTrack+1);
						if (Track == StartTrack+1) Gap[DestDrive[i-1]] = GapLength(VB[DestNumber%2 ? (i+Track+1)%2 : (i-1)%2]);
						VerifyAndRetry(DestDrive[i-1], Track-1, VB[DestNumber%2 ? (i+Track+1)%2 : (i-1)%2]);
						if (Track == 81 && MultipleRootBlocks && IsDos) {
							UpdateMFMRootBlock(B[80%3], Date, IncName);
							StartWrite(MASK(DestDrive[i]), 80, B[80%3]-WriteGap, WRITELENGTH+WriteGap, DestDrive[i], VB[DestNumber%2 ? (i+1)%2 : i%2], TRUE);
						}
						WaitDMA();
						if (Track<=EndTrack) MoveHead(DestDrive[i], Track);
					}
					if (Track>EndTrack) VerifyAndRetry(DestDrive[DestNumber-1], EndTrack, VB[DestNumber%2 ? Track%2 : 1]);
				}
				else {
					if (Track != 81 || !MultipleRootBlocks || !IsDos) StartWrite(ActualDest, Track-1, B[(Track-1)%3]-WriteGap, WRITELENGTH+WriteGap, DestDrive[0], (Track == StartTrack+1) ? VB[0] : NULL, TRUE);
					else for(i=0; i<DestNumber; i++) {
						WaitDMA();
						UpdateMFMRootBlock(B[80%3], Date, IncName);
						StartWrite(MASK(DestDrive[i]), 80, B[80%3]-WriteGap, WRITELENGTH+WriteGap, DestDrive[0], (i == 0 && Track == StartTrack+1) ? VB[0] : NULL, TRUE);
					}
					DrawBar(Track);
					if (Track<=EndTrack) {
						if (Formatting) CreateTrack(Track, B[Track%3], UseFFS);
						else {
							if (Compression) {
								if (Track%2) {
									WaitCompression(Track == EndTrack);
									if (Track != EndTrack) StartDecompression(Track+1, B[(Track+1)%3], B[(Track+2)%3]);
								}
								else WaitInterDecompression();
							}
							RetrieveAndCheck(Track);
						}
					}
					if (Track == StartTrack+1) {
						for(i=1; i<DestNumber; i++) {
							StartRead(DestDrive[i], Track-1, VB[i%2], TRUE);
							Gap[DestDrive[i-1]] = GapLength(VB[(i-1)%2]);
						}
						WaitDMA();
						Gap[DestDrive[i-1]] = GapLength(VB[(i-1)%2]);
						WriteGap = 0;
						for(i=0; i<DestNumber; i++) if (Gap[DestDrive[i]]>WriteGap) WriteGap = Gap[DestDrive[i]];
					}
					if (Track<=EndTrack) MoveHeads(ActualDest, Track);
				}
				if (Track == 1 && (Buffer == VDISK_BUFFER || Buffer == HD_BUFFER) && !Formatting) BufferIsDos = IsDos = IsDosBuffered();
			}
			Track++;
		}

		WaitDMA();
		WaitHeads();
		StopMotorsInUse();
		CopyTaskGetSystemTime(&b);
		WriteElapsed(&a, &b);
		if (Buffer == VDISK_BUFFER) CopyTaskCloseVDisk();
		if (Buffer == HD_BUFFER) CopyTaskCloseHDBuffer();


		if (!Stopped) {
			if (Copying == 'R') {
				ValidBuffer = TRUE;
				if (Compression) {
					if (WaitCompression(TRUE)) {
						BET = EndTrack-2;
						WipeProgress(EndTrack/2);
					}
					else BET = EndTrack;
				}
				if (StartTrack == BST) {
					Copies = 0;
					WriteCopies();
				}
			}
			Beep(4000, TRUE);
			if (Buffer && !Formatting && Copying == 'G' && StartTrack == BST && EndTrack == 159) {
				for(i=0; i<4; i++) if (!Errors[i] && INACTUALDEST(i)) Copies++;
				WriteCopies();
			}
		}

		if (Compression) WaitCompression(TRUE);
		Signal(Me, CopierMask);
	}

	UnLock(CurrentDir(CDLock));
	CopyTaskCloseTimer();
	CopyTaskFreeStuff();
	CloseAudio();
	Signal(Me, CopierMask);
}

/*
 * The big thing!
 * Apart from the open-all-stuff/close-all-stuff parts, this is simply a
 * preposterously huge event loop.
 */

void _main(char *line) {

	register unsigned char i;
	unsigned char *p;

	struct Gadget *G;
	struct IntuiMessage *Msg;
	ULONG Class, Code, Qualifier;
	short Op, OldPri;

	stdin = Input();
	stdout = Output();

	if (ParseCommandLine(line) || FindTask(CopyTaskName) || FindPort(ARexxPortName)) return;

	if (GfxBase = (void *)OpenLibrary(GfxName, 36)) NewOS = TRUE;
	else if ((GfxBase = (void *)OpenLibrary(GfxName, 33)) == NULL) return;

	if ((OldPri = SetTaskPri(Me = FindTask(NULL), 2))>2) SetTaskPri(Me,OldPri);

	SetUpImages();

	if (!AllocStuff() || (DiskBase = OpenResource(DISKNAME)) == NULL ||
		!CreateProc(CopyTaskName, Me->tc_Node.ln_Pri+1, MKBADDR(FakeSegList), 4000)) goto GameOver;
	Wait(CopierMask);
	if (Done) goto GameOver;
	DRU.dru_Message.mn_ReplyPort = ResPort;
	if (!ARexxPort && !GUI) goto GameOver;
	EnableGadgetMask((PresentDisks = WhichDisks()) | (ARexx ? MASK(AREXX) : 0));

	if (!OpenTimer() ||
		(IntuitionBase = (void *)OpenLibrary("intuition.library", 33)) == NULL ||
		!SetUpWindow())
		goto GameOver;

	StarTrek();

	B[3] = VB[0] = VerifyBuffer[0]+8;
	B[4] = VB[1] = VerifyBuffer[1]+8;
	if (GUI && ARexxPort && StartARexxMacro(CreateArgstring(StartupName, strlen(StartupName)+1), TRUE)) PendingMacros = 1;

	PostClick();

	FOREVER {

		do {
			if (Signals & CopierMask) {
				Signals &= ~CopierMask;
				FlushVars();
				AfterCopy();
			}

			if (!Copying && CheckClick()) {
				if (Dest || Source>=0) {
					RemoveClick();
					for(i=0; i<4; i++)
						if (INMASK(i,PresentDisks)) {
							if (GetCIAAPRA(i, CIAF_DSKMOTOR, TRUE) & CIAF_DSKCHANGE) {
								DisksIn |= MASK(i);
							}
							else {
								if (DisksIn & MASK(i)) {
									TimerWait(100000);
									RecalibrateHead(i);
								}
								DisksOut |= MASK(i);
								DisksIn &= ~MASK(i);
							}
						}
					i = Dest | (Source<0 || Buffer || Formatting ? 0 : MASK(Source));
					if (Dest && ((Source>=0 && !INDEST(Source)) || Buffer || Formatting) && i && (DisksIn & i) == i && (DisksOut & i) == i) AutoFlag = Auto;
					ClickHeads(PresentDisks, ++ClickCount);
					PostClick();
				}
			}

			if ((Window && (Msg = (void *)GetMsg(Window->UserPort))) || AutoFlag) {

				if (Msg) {
					Qualifier = Msg->Qualifier;
					G = (struct Gadget *)(Msg->IAddress);
					if ((Class = Msg->Class) == RAWKEY) Code = DeadKeyConvert(Msg);
					else Code = Msg->Code ;
					ReplyMsg((void *)Msg) ;
				}
				else {
					Code = 'G';
					Class = RAWKEY;
				}

				switch(Class) {
					case CLOSEWINDOW: goto GameOver;
											break;

					case REFRESHWINDOW:	RefreshWindow();
												break;

					case GADGETUP:
					case GADGETDOWN:
					case RAWKEY:
						if (Class == RAWKEY) {
							if (Code == 0x5F) {
								HyperHelp();
								break;
							}
							if (Msg && ((Qualifier & IEQUALIFIER_REPEAT) || !Code)) break;
							G = GAddr(Op2Gadget(Op = toupper(Code)));
						}
						else G = GAddr(Op2Gadget(Op = G->GadgetID));

						if (!ShiftedKey(Op, Class, G)) switch(Op) {
							case 'C':
							case 'R':
							case 'G':	SwitchActions(Op);
											break;

							case 'X':	if (!G) break;
											if (StartARexxMacro(GetFile(), FALSE)) PendingMacros++;
											break;
							case ESC_CODE:
							case 'Q':	goto GameOver;
											break;

							case 'S':	Stop();
											break;

							case 'E':
							case 'K':	if (KillSys(Op)) goto GameOver;
											break;

							case 'I':
							case 'F':
							case 'H':
							case 'A':
							case 'U':
							case 'B':
							case 'V':
							case 'D':
							case '.':
							case 'T':	if (!G) break;
											if (Class == RAWKEY) ToggleGadget(G);
											i =  ((G->Flags & SELECTED) != 0);
											SwitchOp(Op, i);
											break;

							case 'Y':
							case 'M':
							case 'L':	ActivateStringGadget(Op);
											break;

							default:		break;
						}
						break;
				}
			}

			if (ARexxPort && (RemRexx = RexxMsg = (void *)GetMsg(ARexxPort))) {
				if (RexxMsg->rm_Node.mn_Node.ln_Type == NT_REPLYMSG) {
					AfterARexxMacro(RexxMsg);
					RexxMsg = NULL;
					if (!--PendingMacros && !NewOS) CloseARexxConsole();
				}
				else {
					SetError(RexxMsg, 0);
					if ((Op = ParseARexx(RexxMsg, &p)) == 0) SetError(RexxMsg, 10);
					else {
						switch(Op) {
							case 'C':
							case 'R':
							case 'G':	if (!PendingRexxMsg) {
												if (!(i = SwitchActions(Op))) PendingRexxMsg = RexxMsg;
												else SetError(RexxMsg, i);
											}
											else SetError(RexxMsg, 1);
											break;

							case 'Q':	goto GameOver;
											break;

							case 'S':	Stop();
											break;
							case 'E':
							case 'K':	if (KillSys(Op)) {
												SetError(RexxMsg,	20);
												goto GameOver;
											}
											break;
							case 9:
							case 11:
							case 'I':
							case 'F':
							case 'H':
							case 'A':
							case 'U':
							case 'B':
							case 'V':
							case 'D':
							case '.':
							case 'T':	if ((i = SwitchOn(p)) > 1) SetError(RexxMsg, 10);
											else if (Op<' ') {
												switch(Op) {
													case 9:	Requesters = i;
																break;
													case 11:	if (PrintErrors = i) OpenErrorConsole();
																else CloseErrorConsole();
																break;
												}
											}
											else if (!(G = GAddr(Op2Gadget(Op)))) SetError(RexxMsg, 1);
											else {
												SelectGadget(G, i);
												SwitchOp(Op, i);
											}
											break;
							case 1:	if (SetDiskName(p)) SetError(RexxMsg, 1);
										break;
							case 2:	if (SetVDNumber(atoi(p))) SetError(RexxMsg, 1);
										break;
							case 3:	if (SetDeviceName(p)) SetError(RexxMsg, 1);
										break;
							case 4:	if (SetRetryNumber(atoi(p))) SetError(RexxMsg, 1);
										break;
							case 5:	if (SetSCyl(atoi(p))) SetError(RexxMsg, 1);
										break;
							case 6:	if (SetECyl(atoi(p))) SetError(RexxMsg, 1);
										break;
							case 7:	if (Copying) {
											SetError(RexxMsg, 1);
											break;
										}
										if (!SwitchOn(p)) {
											if (Source>=0) SelectGadget(GAddr(DF0S)+Source, FALSE);
											Source = -1;
										}
										else {
											Op = atoi(p);
											if (Op>3 || Op<0) {
												SetError(RexxMsg,10);
												break;
											}
											if (MASK(Op) & ~PresentDisks) {
												SetError(RexxMsg,2);
												break;
											}
											if (Source == Op) break;
											if (Source>=0) SelectGadget(GAddr(DF0S)+Source, FALSE);
											if (GrabUnits(PresentDisks)) {
												Source = Op;
												SelectGadget(GAddr(DF0S)+Op, TRUE);
											}
											else {
												Source = -1;
												SetError(RexxMsg, 5);
											}
										}
										if (!Dest && Source<0) DropUnits(PresentDisks);
										ResetAuto();
										break;
							case 8:	if (Copying) {
											SetError(RexxMsg, 1);
											break;
										}
										Op = atoi(p);
										if (Op > 15 || Op <0) {
											SetError(RexxMsg, 10);
											break;
										}
										if (Op & ~PresentDisks) {
											SetError(RexxMsg, 2);
											break;
										}
										for(i=0; i<4; i++) SelectGadget(GAddr(DF0D)+i, INMASK(i,Op) != 0);
										if (Op && !GrabUnits(PresentDisks)) {
											DeselectDiskGadgets();
											Dest = 0;
											SetError(RexxMsg, 5);
										}
										else Dest = Op;
										if (!Dest && Source<0) DropUnits(PresentDisks);
										ResetAuto();
										break;
							case 10:	if (SetVolume(atoi(p))) SetError(RexxMsg, 1);
										break;
							case 13:	if ((NoClick = atoi(p)) & ~0xF) {
											NoClick = 0;
											SetError(RexxMsg, 1);
										}
										break;
							case 12: if (StartARexxMacro(CreateArgstring(p, strlen(p)+1), FALSE)) PendingMacros++;
										else SetError(RexxMsg, 1);
											break;
						}

					}
					if (PendingRexxMsg != RexxMsg) ReplyMsg((void *)RexxMsg);
					RexxMsg = NULL;
				}
			}

		} while((Window && Msg) || (ARexxPort && RemRexx) || ((Dest || Source>=0) && !Copying && CheckClick()) || AutoFlag);

		Signals = Wait(CopierMask | (Window ? MASK(Window->UserPort->mp_SigBit) : 0) | MASK(ClickPort->mp_SigBit) | (ARexxPort ? MASK(ARexxPort->mp_SigBit) : 0));
	}

GameOver:

	Done = TRUE;
	if (Copying) {
		Signal(Copier, StopMask);
		Wait(CopierMask);
	}
	if (Copier) {
		Signal(Copier, StartMask);
		Wait(CopierMask);
	}
	if (PendingRexxMsg) {
		SetError(PendingRexxMsg, 30);
		ReplyMsg(PendingRexxMsg);
	}
	while (PendingMacros) {
		WaitPort(ARexxPort);
		while((RemRexx = (void *)GetMsg(ARexxPort)) && RemRexx->rm_Node.mn_Node.ln_Type != NT_REPLYMSG) {
			SetError(RemRexx, 30);
			ReplyMsg(RemRexx);
		}
		if (RemRexx) {
			AfterARexxMacro(RemRexx);
			PendingMacros--;
		}
	}
	FreeAllBuffers();
	CloseHyperHelp();
	RemovePostedDRU();
	DropUnits(PresentDisks);
	SetTaskPri(Me, OldPri);
	RemoveClick();
	MyCloseWindow();
	ShutScreen();
	CloseErrorConsole();
	if (IntuitionBase) CloseLibrary((void *)IntuitionBase);
	if (GfxBase) CloseLibrary((void *)GfxBase);
	CloseVoice();
	CloseTimer();
	if (RexxMsg) ReplyMsg(RexxMsg);
	FreeStuff();
	CloseARexxConsole();
}
