#include "SD.h"
#include <proto/console.h>

/* NOTE: apart from the CopyTask*() functions, all this module has to be used
	only from the main task. */

ULONG StepDelay;
ULONG SettleDelay;
ULONG CalibrateDelay = 4000;

extern struct Library *AslBase;
struct Library *ConsoleDevice;

unsigned char	DefaultDir[] = "REXX:#?.supdup",
					DefaultTitle[] = "Select an ARexx program",
					ARexxPortName[] = "SUPERDUPER",
					ShiftedKeyCodes[4],
					UnshiftedKeyCodes[4];

struct MsgPort *TPort,			/* For all temporary (sync) IOs */
					*ARexxPort, 	/* For ARexx */
					*ClickPort, 	/* For the click timings */
					*ResPort;		/* For the resource messages */

struct Task *Me, *Copier;

static unsigned char	NoDiskResourceMsg[] = "Can't get disk.resource.";
static unsigned char	FreeDiskResMsg[] = "Please free disk.resource.";

static unsigned short OriginalCyl[4];  /* The cylinder position, when we grabbed it */
static struct IOStdReq *ConsoleIOStdReq;

void MemCleanup(void) {
}	/* we don't need it */

/* This are all signals of the copy task */

static signed char	CompSignal = -1,		/* Compression is over signal */
							EBNSignal = -1,		/* Blitter finished signal */
							ISDSignal = -1,		/* Disk read/write finished signal */
							StartSignal = -1,		/* "Start the copy" signal */
							StopSignal = -1;		/* "Stop the copy" signal */

/* This are all signals of the main task */

static signed char	CopierSignal = -1;		/* When the copy task has something to say */

static unsigned char GotUnit,			/* I *have* the disk.resource */
							InhibitMask,	/* What drives did I inhibit? */
							DRUPosted;		/* There's a DRU posted with no reply still hanging */



void FlushMem(void) {

	void *p;
	if (p = AllocMem(0x80000000UL, MEMF_PUBLIC)) FreeMem(p, 0x80000000UL);
	if (p = AllocMem(0x80000000UL, MEMF_PUBLIC)) FreeMem(p, 0x80000000UL);
}

/* Frees all ARexx-related libraries. */


static void CloseARexxStuff(void) {

	if (RexxSysBase) CloseLibrary((void *)RexxSysBase);

	if (AslBase) {
		ASLFreeFileReq();
		CloseLibrary(AslBase);
	}

	if (ArpBase) CloseLibrary(ArpBase);

	if (ReqBase) {
		REQFreeFileReq();
		CloseLibrary(ReqBase);
	}
}

/* We ask for a file name. We return NULL if nothing was chosen, otherwise
	a pointer to an ArgString. */

unsigned char *GetFile(void) {
	if (AslBase) return(ASLGetFile());
	else if (ArpBase) return(ARPGetFile());
	else if (ReqBase) return(REQGetFile());
	else return(NULL);
}

/* Here we feed the console device with fake InputEvent in order to
	discover what '0', '1',..., SHIFT-'0', SHIFT-'1' etc. is. We put the result in
	ShiftedKeyCodes and UnshiftedKeyCodes. We use TPort. ConsoleDevice must be opened. */

static void CalcKeyCodes(void) {

	struct InputEvent ievent = { NULL, IECLASS_RAWKEY };
	unsigned char c;
	unsigned int i;

	ievent.ie_position.ie_addr = NULL;

	for(i=0; i<4; i++) {
		ievent.ie_Code = (i ? i : 10);
		ievent.ie_Qualifier = (IEQUALIFIER_LSHIFT|IEQUALIFIER_RSHIFT);
		if (RawKeyConvert(&ievent, &c, 1, NULL)>0) GAddr(DF0D+i)->GadgetID = ShiftedKeyCodes[i] = c;
		ievent.ie_Qualifier = 0;
		if (RawKeyConvert(&ievent, &c, 1, NULL)>0) GAddr(DF0S+i)->GadgetID = UnshiftedKeyCodes[i] = c;
	}
}

unsigned char DeadKeyConvert(struct IntuiMessage *msg) {

	unsigned char c;
	static struct InputEvent ievent = { NULL, IECLASS_RAWKEY };

	ievent.ie_Code = msg->Code;
	ievent.ie_Qualifier = msg->Qualifier;
	ievent.ie_position.ie_addr = *((APTR *)msg->IAddress);
	if (RawKeyConvert(&ievent, &c, 1, NULL)>0) return(c);
	else return(0);
}


/* Main allocation function. It allocates everything. If ARexx is set to one, tries to
	open ARexx. If for any reason (no ARexx or no file requester) the gadget ARexx should
	be ghosted, it sets ARexx to 0. Check ARexxPort to see if you should use ARexx,
	the ARexx variable to see if you should activate the gadget (dirty and twisted, I agree).
	Moreover, the SignalSemaphore for sharing the video data is initialized. */

int AllocStuff(void) {

	register unsigned char i;

	InitSemaphore(&Semaphore);
	InitSemaphore(&HyperSemaphore);

	if ((CopierSignal = AllocSignal(-1)) != -1) CopierMask = MASK(CopierSignal);

	TPort = CreatePort(NULL,0);
	ResPort = CreatePort(NULL,0);
	ClickPort = CreatePort(NULL,0);
	if (TPort && (ConsoleIOStdReq = CreateStdIO(TPort)) && !OpenDevice("console.device", -1, ConsoleIOStdReq, 0)) {
		ConsoleDevice = (void *)ConsoleIOStdReq->io_Device;
		CalcKeyCodes();
	}

	if (ARexx) {
		if (!FindPort(ARexxPortName) && (RexxSysBase = (void *)OpenLibrary(RXSNAME, 0)) && (ARexxPort = CreatePort(ARexxPortName,0))) {
			if (GUI && !((NewOS && (AslBase = (void *)OpenLibrary("asl.library", 0))) ||
				(ArpBase = (void *)OpenLibrary("arp.library", 0)) ||
				(ReqBase = (void *)OpenLibrary("req.library", 0))))
				ARexx = 0;
		}
		else ARexx = 0;
	}
	for(i=0; i<3; i++)
		if (ReadBuffer[i] = AllocMem(READBUFFERSIZE, MEMF_CHIP)) setmem(ReadBuffer[i], GAPSIZE+8, 0xAA);
	for(i=0; i<2; i++) VerifyBuffer[i] = AllocMem(VERIFYBUFFERSIZE, MEMF_CHIP);
	return(CopierSignal != -1 && VerifyBuffer[0] && VerifyBuffer[1] && ReadBuffer[0] && ReadBuffer[1] && ReadBuffer[2] && TPort && ClickPort && ResPort && (!RexxSysBase || ARexxPort) && ConsoleDevice);
}

int CopyTaskAllocStuff(void) {

	if ((StopSignal = AllocSignal(-1)) != -1) StopMask = MASK(StopSignal);
	if ((StartSignal = AllocSignal(-1)) != -1) StartMask = MASK(StartSignal);
	if ((CompSignal = AllocSignal(-1)) != -1) CompMask = MASK(CompSignal);
	if ((EBNSignal = AllocSignal(-1)) != -1) EBN.EBN_Mask = MASK(EBNSignal);
	if ((ISDSignal = AllocSignal(-1)) != -1) DiskBlockServerData.DBSD_ISD.ISD_Mask = SoftIntData.SID_ISD.ISD_Mask = MASK(ISDSignal);

	return(StartSignal != -1 && StopSignal != -1 && CompSignal != -1 && EBNSignal != -1 && ISDSignal != -1);
}


/* No comment */

void FreeStuff(void) {

	register unsigned char i;
	struct RexxMsg *ARexxMessage;

	if (CopierSignal != -1) FreeSignal(CopierSignal);

	for(i=0; i<3; i++) if (ReadBuffer[i]) FreeMem(ReadBuffer[i], READBUFFERSIZE);
	for(i=0; i<2; i++) if (VerifyBuffer[i]) FreeMem(VerifyBuffer[i], VERIFYBUFFERSIZE);

	if (ARexxPort) {
		Forbid();
		while(ARexxMessage=(struct RexxMsg *)GetMsg(ARexxPort)) {
			ARexxMessage->rm_Result1 = 30;
			ARexxMessage->rm_Result2 = NULL;
			ReplyMsg((struct Message *)ARexxMessage);
		}
		DeletePort(ARexxPort);
		Permit();
	}

	CloseARexxStuff();
	if (ConsoleDevice) CloseDevice(ConsoleIOStdReq);
	if (ConsoleIOStdReq) DeleteStdIO(ConsoleIOStdReq);
	if (TPort) DeletePort(TPort);
	if (ClickPort) DeletePort(ClickPort);
	if (ResPort) DeletePort(ResPort);
}

void CopyTaskFreeStuff(void) {

	if (StartSignal != -1) FreeSignal(StartSignal);
	if (StopSignal != -1) FreeSignal(StopSignal);
	if (CompSignal != -1) FreeSignal(CompSignal);
	if (EBNSignal != -1) FreeSignal(EBNSignal);
	if (ISDSignal != -1) FreeSignal(ISDSignal);
}



/*
 *	  This routine gives us those pretty little DF0:BUSY things, which keep
 *	  AmigaDOS from futzing with the drives when we are playing with them.
 *	  Uses TPort.
 */

static void InhibitDisk(unsigned char unit, unsigned char bool) {

	register struct MsgPort *handler;
	static struct StandardPacket packet;

	if (unit>3) return;
	handler = (struct MsgPort *)DeviceProc(df[unit]);
	if (handler == NULL) return;

	packet.sp_Msg.mn_Node.ln_Name = (char *)&(packet.sp_Pkt);
	packet.sp_Pkt.dp_Link = &(packet.sp_Msg);
	packet.sp_Pkt.dp_Port = TPort;
	packet.sp_Pkt.dp_Type = ACTION_INHIBIT;
	packet.sp_Pkt.dp_Arg1 = bool;
	PutMsg(handler, (void *)&packet);
	WaitPort(TPort);
	GetMsg(TPort);
}



/*
 *	  Now we try to allocate a disk.	 First, we attempt to allocate an
 *	  I/O request, and then we actually open the device.	If either of
 *	  these fail, we return 0.	 This can occur if a drive is opened that
 *	  doesn't exist. The drive will get TPort as its I/O port.
 */
static struct IOStdReq *OpenDisk(unsigned char Unit) {

	register struct IOStdReq *IOStdReq;

	if ((IOStdReq = CreateStdIO(TPort)) && OpenDevice(TD_NAME, Unit, IOStdReq, 0) == 0) return(IOStdReq);
	if (IOStdReq) DeleteStdIO(IOStdReq);
	return(NULL);
}

/*
 *	  Here we release a disk.	We close the device, kill the I/O request, and exit.
 */
static void CloseDisk(struct IOStdReq *IOStdReq) {
	CloseDevice(IOStdReq);
	DeleteStdIO(IOStdReq);
}

/*
 * We check which disks are present. We return an ULONG which we will OR
 * with the EnableGadgets standard mask. We use TPort. We return 0 if
 * the unit 0 can't be opened.
 */

unsigned char WhichDisks(void) {

	register unsigned char i, result = 0;
	register struct IOStdReq *IOStdReq;

	for(i=0; i<4; i++)
		if (GetUnitID(i) == DRT_AMIGA) {
			if (IOStdReq = OpenDisk(i)) {
				TDU_PublicUnit[i] = (struct TDU_PublicUnit *)IOStdReq->io_Unit;
				StepDelay = max(StepDelay, TDU_PublicUnit[i]->tdu_StepDelay);
				SettleDelay = max(SettleDelay, TDU_PublicUnit[i]->tdu_SettleDelay);
				if (NewOS) CalibrateDelay = max(CalibrateDelay, TDU_PublicUnit[i]->tdu_CalibrateDelay);
				CloseDisk(IOStdReq);
			}
			result |= 0x11<<i;
		}
	if (result & 1) return(result);
	else return(0);
}



/* Here we remove the posted DRU. Until it arrives, the program is
	forced to hang! */

void RemovePostedDRU(void) {
	if (DRUPosted) {
		while(!GetMsg(ResPort)) Acknowledge(FreeDiskResMsg);
		DRUPosted = 0;
	}
}

/* Here we grab the units, stealing them from the system. We recalibrate
	the heads and we store the original track, so we can get out of here
	under 1.3 without asking for ejection of all disks. */

unsigned char GrabUnits(unsigned char UnitMask) {

	unsigned char Unit;

	if (GotUnit) return(1);

	if (DRUPosted) {
		if (GetMsg(ResPort)) DRUPosted = 0;
		else {
			Acknowledge(NoDiskResourceMsg);
			return(0);
		}
	}

	UnitMask &= 0xF;

	for(Unit=0; Unit<4; Unit++)
		if (TDU_PublicUnit[Unit]) {
			InhibitDisk(Unit,TRUE);
			InhibitMask |= MASK(Unit);
		}

	if (!GetUnit(&DRU)) {
		DropUnits(UnitMask);
		Acknowledge(NoDiskResourceMsg);
		DRUPosted = 1;
		return(0);
	}

	GotUnit = 1;
	for(Unit=0; Unit<4; Unit++) if (INMASK(Unit,UnitMask)) OriginalCyl[Unit] = FindTrack(Unit);
	return(1);
}

/* We release the units to the system. First of all, we restore the heads to the
	original position. */

void DropUnits(unsigned char UnitMask) {

	unsigned char Unit;

	UnitMask &= 0xF;
	if (GotUnit) {
		for(Unit=0; Unit<4; Unit++)
			if (TDU_PublicUnit[Unit]) {
				if (NewOS) SetPublicUnit(Unit);
				else {
					ResetHead(Unit, OriginalCyl[Unit]*2);
					MotorOff(MASK(Unit));
				}
			}
		GiveUnit();
		GotUnit = 0;
	}
	if (InhibitMask) {
		for(Unit=0; Unit<4; Unit++) if (INMASK(Unit,InhibitMask)) InhibitDisk(Unit,FALSE);
		InhibitMask = 0;
	}
}
